#-*- coding: utf-8 -*-
"""This file contains constants which represent physical properties and
scientific principles which will be used in AguaClara plant design.

"""
from aguaclara.core.units import unit_registry as u
import aguaclara.core.pipedatabase as pipe
import numpy as np

#####tabulated constants

GRAVITY = 9.80665 * u.m/u.s**2

#Density of water
DENSITY_WATER = 1000 * u.kg/u.m**3

#The kinematic viscosity of water
NU_WATER = 1 * 10**-6 * u.m**2/u.s

#The influence of viscosity on mixing in jet reactors
RATIO_JET_ROUND = 0.5

#This is an estimate for plane jets as created in the flocculator and
# in the sed tank jet reverser.
RATIO_JET_PLANE = 0.225

RATIO_VC_ORIFICE = 0.63

P_ATM = 1*u.atm

#Needed for the filter siphon design
NU_AIR = 12 *u.mm**2/u.s

#Needed for the filter siphon design
RHO_AIR = 1.204 * u.kg/u.mm**3

####General assumptions and constants

PLANT_ORIGIN=[0, 0, 0] * u.m

COUNTRY = 0

LANGUAGE = 0

#Prompts the transition to a low flow plant
FLOW_PLANT_MAX_LF = 16.1 * u.L/u.s

WIDTH_HUMAN_MIN = 0.5 * u.m

#The height of the walkway above the drain channel bottom so that
# someone can walk through the drain channel.
HEIGHT_HUMAN_ACCESS = 1.5 * u.m

#Used to set the minimum height of entrance, floc, and sed walls
HEIGHT_PLANT_FREE_BOARD = 0.1 * u.m

#Minimum space between fittings in a tank or fittings and the wall of
# the tank.
SPACE_FITTING = 5 * u.cm

#Minimum channel width for constructability
WIDTH_CHANNEL_MIN = 15 * u.cm

#RATIO_RECTANGULAR is defined as the optimum "height over width ratio"
# (1/2) for a rectangular open channel
RATIO_RECTANGULAR = 0.5

##Equals 1 to draw boxes showing max water levels, 0 normally
#EN_WATER=0 #ASK Monroe

##If EN_WATER is set to 1, this controls the filter operation mode for
# which water/sand elevations are drawn. 0 for terminal, 1 for clean bed,
# and 2 for backwash.
#EN_WATER=2 #ASK Monroe

WIDTH_DOOR = 1 * u.m

THICKNESS_ACRYLIC = 1 * u.cm

#Due to a 24 in LFOM because that's the biggest pipe we have in our
# database right now. if we need a bigger single train, we can do that
# by adding that pipe size into the pipe database
FLOW_TRAIN_MAX = 150.1 * u.L/u.s

def en_multiple_train(flow_plant):
    if flow_plant > 60 * u.L/u.s:
        return 1
    else:
        return 0


def n_train(flow_plant):
    if flow_plant > 60*u.L/u.s:
        return 1
    elif flow_plant > 60*u.L/u.s and flow_plant <= 120 * u.L/u.s:
        return 2
    else:
        return 4


def flow_train(flow_plant):
    return flow_plant / n_train(flow_plant)

#####Flow orifice meter

#Maximum number of rows or orifices in lfom.
RATIO_LFOM_ORIFICE = 10

#Safety coefficient that ensures free fall at the bottom of the lfom pipe
RATIO_LFOM_SAFETY = 1.5

#Minimum safety coefficient for lfom pipe diameter; only reduced
# between 55 L/s and 70 L/s - the intermediate zone between the using an
# LFOM pipe and an LFOM channel.
#It may be possible to eliminate this if we switch to plate LFOM at 50 Lp
RATIO_LFOM_SAFETY_MIN = 1.15

#Minimum head loss through linear orifice meter
HEADLOSS_LFOM_MIN = 20 * u.cm

#Maximum head loss through linear orifice meter to be used as needed for high flow plants.
HEADLOSS_LFOM_MAX = 40 * u.cm

#Changed from 12 in by pc479 because this is not a constraint anymore
# because we don't have an elbow constraining us. LFOM still needs to fit
# in the entrance tank. Need to check this constraint (mrf222)
NOM_DIAM_LFOM_PIPE_MAX = 36 * u.inch

HEIGHT_LFOM_FREEFALL = 10 * u.cm

#Enumerated type that selects between pipe(1) and plate(0) for the LFOM
def en_lfom_pipe(flow_plant):
    if flow_plant >= 80 * u.L/u.s:
        return 0
    else:
        return 1

####entrance tank

#Used to make a smaller entrance tank if the source water doesn't contain grit.
##0 if we want entrance tank sized to capture grit, 1 for minimum size
##EN_GRIT=0    ASK Monroe

##0 if there is only one inlet, 1 if there is two inlet
##EN_TWO_INLETS=0

#Angle of the sloped walls of the entrance tank hoppers
ANGLE_ENT_TANK_SLOPE = 45 * u.deg

#Extra space around the float (increase in effective diameter) to ensure
# that float has free travel
SPACE_ENT_TANK_FLOAT = 5 * u.cm

#Increased to get better mixing (10/10/2015 by Monroe)
ENERGY_DIS_RAPID_MIX = 3 * u.W/u.kg

#Distance that the rapid mix coupling extends into the first floc channel
# so that the RM orifice place can be fixed in place.
LENGTH_FLOC_COUPLING_EXT = 5 * u.cm

WIDTH_ENT_TANK_HOPPER_PEAK = 3 * u.cm

#Distance from the front wall to the pipe stubs in the hopper drains so
# that an operator can easily reach them.
LENGTH_ENT_TANK_WALLTODRAIN_MAX = 40 * u.cm

#Entrance tank capture velocity
VEL_ENT_TANK_CAPTURE_BOD = 8 * u.mm/u.s

AN_ENT_TANK_PLATE = 50 * u.deg

SPACE_ENT_TANK_PLATE = 2.5 * u.cm

THICKNESS_ENT_TANK_PLATE = 2 * u.mm

DIST_CENTER_ENT_TANK_PLATE = SPACE_ENT_TANK_PLATE + THICKNESS_ENT_TANK_PLATE

NOM_DIAM_ENT_TANK_MOD = 0.5 * u.inch

NOM_DIAM_ENT_TANK_MOD_SPACER = 0.75 * u.inch

#Thickness of the PVC disk used as the float for the chemical dose
# controller lever arm.
THICKNESS_ENT_TANK_FLOAT = 5 * u.cm

SPACE_ENT_TANK_LAMINA_PIPETOEDGE = 5 * u.cm

NOM_DIAM_RAPID_MIX_PLATE_RESTRAINER = 0.5 * u.inch

#Nom diam of the pipes that are embedded in the entrance tank slope
# to support the plate settler module
NOM_DIAM_ENT_TANK_PLATE_SUPPORT = 3 * u.inch

####chemical dose controller

##0 is alum, 1 is PACl
##EN_COAG=1

M_COAG_SACK = 25 * u.kg

#The coagulant stock is relatively stable and can last many days. Here we
# set the minimum time the coagulant stock will last when applying the
# maximum possible dose to size the stock tanks. In general the dose will
# be less than this and the stock will last much longer.
TIME_COAG_STOCK_MIN_EST = 1 * u.day

#Want chlorine stock to run out on average every day so that the stock
# is made fresh frequently because the chlorine stock degrades with time
# depending on temperature, concentration, and pH.
TIME_CHLOR_STOCK_AVE = 1 * u.day

ID_COAG_TUBE = 0.125 * u.inch
#1/8" tubes are readily available in hardware stores in Honduras
ID_CHLOR_TUBE = 0.125 * u.inch

CONC_COAG_STOCK_EST = 150 * u.g/u.L

CONC_CHLOR_STOCK_EST1 = 15 * u.g/u.L

P_CHLOR = 0.7

#This is the elevation difference between the outlet of the coagulant
# stock tanks and the water level in the constant head tank, which is set
# by the hydraulic head required to provide the desired max chemical flow
# rate through the float valve orifice in the CHT.
#It is treated as constant here to ensure a practical elevation difference
# is left between the stock tanks and the CHT even when a float valve is
# selected which requires very little hydraulic head to deliver the
# required maximum chemical flow rate.
HEIGHT_COAG_TANK_ABOVE_HEAD_TANK = 30 * u.cm

#This is the distance from the bottom of the stock tanks to the outlets
# to allow space for solids to settle.
DIST_CENTER_STOCK_OUTLET = 10 * u.cm

#Distance between a tank and the border of the platform
SPACE_CHEM_TANK_BORDER = 5 * u.cm

#This is the estimated elevation difference between the water level in
# the constant head tank and the top of the entrance tank wall.
#The constant head tank water level is the same as the elevation of the
# outlet of the dosing tube when the lever arm is horizontal (zero flow).
#Therefore this height depends only on the hardware used to make the
# slider/drop tube assembly and to mount the lever arm to the entrance
# tank wall.
#Note that this will vary depending on hardware used, and is only
# defined here to calculate the elevation of the stock tanks, which can
# be approximate.
HEIGHT_DOSER_ASSEMBLY = 6.77 * u.cm

#Maximum error allowed between a linear flow vs tube head loss
# relationship and the actual performance (which is affected by non-linear
# minor losses), assuming calibration at the maximum flow rate.
RATIO_LINEAR_CDC_ERROR = 0.1

#Estimated minor loss coefficient for the small-diameter flexible tubing
# using fittings that have larger ID than the tubing.
K_MINOR_CDC_TUBE = 2

#Head loss through the doser at maximum flow rate.
#Maximum head loss through the small-diameter dosing tubing, which
# corresponds to the variation in water levels in the entrance tank and
# the difference between the maximum and minimum elevation of the dosing
# tube outlet attached to the lever arm.
HEADLOSS_CDC = 20 * u.cm

#Estimated distance between fluid level in constant head tank and float valve orifice
HEIGHT_CDC_FLOAT_VALVE = 5 * u.cm

#Nominal diameter of the PVC plumbing for the chlorine dosing system.
NOM_DIAM_CHLOR_PIPE = 0.5 * u.inch

#Nominal diameter of the PVC plumbing for the coagulant dosing system.
NOM_DIAM_COAG_PIPE = 0.5 * u.inch

#Supplier Information:
# http://www.rotoplas.com/assets/files/industria/catalogo.pdf
#5-gallon bucket
# http://www.mcmaster.com/#storage-buckets/=kd23oh
#35-gallon drum
# http://www.jlmovingsupplies.com/c31/DIXIE-OPEN-CLOSED-HEAD-DRUMS-p36721.html
VOL_CHEM_TANK_AVAIL = [5 * u.gal, 35 * u.gal, 55 * u.gal,
                       450 * u.L, 750 * u.L, 1100 * u.L, 2500 * u.L]

DIAM__CHEM_TANK_AVAIL = [11.875 * u.inch, 20.75 * u.inch, 22.5 * u.inch,
                         0.85 * u.m, 1.10 * u.m, 1.10 * u.m, 1.55 * u.m]

HEIGHT_CHEM_TANK_AVAIL = [17.75 * u.inch, 31.75 * u.inch, 33.5 * u.inch,
                          0.99 * u.inch, 1.02 * u.inch, 1.39 * u.inch,
                          1.65 * u.inch]

####Chemical dose controller dimensions (based on inserted drawings)

#st587 addition
LENGTH_CDC_LEVER_ARM = 0.5 * u.m

DIAM_CDC_LEVER_CYLINDER1 = 1 * u.inch

DIAM_CDC_LEVER_CYLINDER4 = 2 * u.inch

DIAM_CDC_LEVER_CYLINDER_2 = 0.5 * u.inch

LENGTH_CDC_LEVER_PIVOTTO_CYLINDER2 = 6 * u.cm

LENGTH_CDC_LEVER_CYLINDER_2TO3 = 9.5 * u.cm

LENGTH_CDC_LEVER_PIVOT_BOX = 2 * u.inch

WIDTH_CDC_LEVER_PIVOT_BOX = 1 * u.inch

HEIGHT_CDC_LEVER_PIVOT_BOX = 1 * u.inch

THICKNESS_CDC_LEVER_ARM = 0.125 * u.inch

HEIGHT_CDC_LEVER_ARM = 1 * u.inch

LENGTH_CDC_LEVER_INNERBAR = 7 * u.inch

LENGTH_CDC_LEVER_MOUNTING_PLATE = 6 * u.inch

WIDTH_CDC_LEVER_MOUNTING_PLATE = 0.5 * u.cm

HEIGHT_CDC_LEVER_MOUNTING_PLATE = 2 * u.inch

SPACE_LEVER_TO_ENT_TANK_Z_TOP = 1 * u.cm

THICKNESS_CDC_FLOAT = 5 * u.cm

DIAM_CDC_FLOAT_CABLE = 0.5 * u.cm

LENGTH_CDC_LEVER_SLIDER_ORIGIN_TO_SCREW = 1 * u.inch

THICKNESS_CDC_LEVER_SLIDER = 0.25 * u.inch

HEIGHT_CDC_LEVER_SLIDER = 1.5 * u.inch

LENGTH_CDC_LEVER_SLIDER = 3 * u.inch

HEIGHT_CDC_LEVER_SLIDER_SHORT = 0.125 * u.inch

LENGTH_CDC_LEVER_CYLINDER = 6 * u.inch

LENGTH_ENT_TANK_FRONT_WALL_TO_CDC_FLOAT = 0.874 * u.m

LENGTH_CDC_LEVER = 0.5 * u.m #This may be obsolete now... mrf222 2/10/16

WIDTH_LEVER_ARM = 0.0032 * u.m

HEIGHT_LEVER_ARM = 0.0254 * u.m

DIAM_CDC_CHT = 6 * u.inch

#Distance from the top of the entrance tank to the to the middle of the
# lever arm hole for the cable - (minus the) radius of the hole.
HEIGHT_LEVER_HOLE = 0.0132 * u.m - (0.0095/2 * u.m)

DIAM_CABLE = 0.1 * u.inch

#Edited DLABOrigintoLAOriginZ to accommodate dimensions from McMaster
# vs Inserted Drawing
DIAM_LAB_ORIGIN_TO_LA_ORIGIN_Z = 0.0245 * u.m

#Distance from the lever arm origin to the outside center of the top part
# of the drop tube in the y direction.
LENGTH_LA_ORIGIN_TO_DT_Y = 0.7812 * u.m

#Distance from the lever arm origin to the drop tube in the z direction.
LENGTH_LA_ORIGIN_TO_DT_Z = 0.0429 * u.m

#Distance from the lever arm origin to the center of the drop tube in the
#x direction.
LENGTH_LA_ORIGIN_TO_DT_CENTER_X = 0.0290 * u.m

#Measured from CDC research team's apparatus.
THICKNESS_CDC_REDUCER = 9.5 * u.mm

#Distance from the lever arm origin to the center of the reducer in the
# x direction.
LENGTH_LA_ORIGIN_TO_REDUCER_X = 0.0290 * u.m

#Distance from the lever arm origin to the outside center of the top part
# of the reducer in the y direction.
LENGTH_LA_ORIGIN_TO_REDUCER_Y = 0.7135 * u.cm

#Distance from the lever arm origin to the center of the reducer in the
# x direction.
LENGTH_LA_ORIGIN_TO_REDUCER_CENTER_X = 0.0290 * u.m

#Distance from the lever arm origin to the center of the reducer in the
# y direction.
LENGTH_LA_ORIGIN_TO_REDUCER_CENTER_Y = 0.7919 * u.m

WIDTH_LEVER_BRACKET = 0.625 * u.inch

LENGTH_LEVER_BRAKCET = 1.5 * u.inch

RADIUS_LA_BAR = 0.375 * u.inch

THICKNESS_LEVER_BRACKET = 0.08 * u.inch

DIAM_LA_BAR = 0.375 * u.inch

LENGTH_LA_BAR = 4 * u.inch

LENGTH_SLIDER = 3 * u.inch

WIDTH_SLIDER = 3.2 * 10**-3 * u.m

NOM_DIAM_DROPTUBE = 0.5 * u.inch

HEIGHT_SLIDER = 0.625 * u.inch

LENGTH_DROPTUBE = 0.61 * u.m

#The length of the drop tube needs to be calculated. The drop tube must be
# as long as the supercritical flow.
#Thus the drop tube must extend down to the elevation of the sed tank
# effluent weir. This constant should be removed!

#Outer diameter of fitting- measured from CDC research team's fittin
OUTER_DIAM_CDC_FITTING = 5/32 * u.inch

#Inner diameter of fitting- measured from CDC research team's fitting
ID_CDC_FITTING = 0.126 * u.inch

#Length of fitting - measured from CDC research team's fitting
LENGTH_CDC_FITTING = 0.75 * u.inch

#st587 addition
##Constant Head Tank Dimensions

#five gallons bucket dimensions for constant head tanks

DIAM_CHT = 10 * u.cm

HEIGHT_CHT = 37/3 * u.cm

THICKNESS_CHT_WALL = 1/3 * u.cm

NOM_DIAM_DELIVERY_PIPE = 0.6 * u.inch

PIPE_SCHEDULE_FLEX_TUBE = 2

LENGTH_PVC_BALL_VALVE = 0.1625/4 * u.cm

THICKNESS_MOUNTING_BOARD = 1.5 * u.inch

##Manifold Dimensions

SPACE_CDC_LEVER_TO_MANIFOLD = 40 * u.cm

LENGTH_CHLOR_AIR_RELEASE_PIPE = 30 * u.cm #Arbitratily selected


####Flocculator
##The minor loss coefficient is 2. According to measurements at Agalteca
# and according to
# https://confluence.cornell.edu/display/AGUACLARA/PAHO+Water+Treatment+Publications
# (page 100 in chapter on flocculation)
HEIGHT_FLOC_OPTION = 0

##Increased both to provide a safety margin on flocculator head loss and
# to simultaneously scale back on the actual collision potential we are
# trying to achieve.
K_MINOR_FLOC_BAFFLE = 2.5

SPACE_FLOC_BAFFLE_SET_BACK_PLASTIC= 2 * u.cm

###Target flocculator collision potential basis of design
COLL_POT_FLOC_BOD = 75 * u.m**(2/3)

##Minimum width of flocculator channel required for constructability based
# on the width of the human hip
FLOC_WIDTH_MIN_CONST = 45 * u.cm

##Minimum and maximum distance between expansions to baffle spacing ratio for
#flocculator geometry that will provide optimal efficiency.
RATIO_HS_MIN = 3
RATIO_HS_MAX = 6

##Ratio of the width of the gap between the baffle and the wall and the
# spacing between the baffles.
RATIO_FLOC_BAFFLE = 1

##Max energy dissipation rate in the flocculator, basis of design.
ENERGY_DIS_FLOC_BOD = 10* u.mW/u.kg

TIME_FLOC_DRAIN = 15 * u.min

NOM_DIAM_FLOC_MOD = 0.5 * u.inch

NOM_DIAM_FLOC_SPACER = 0.75 * u.inch

SPACE_FLOC_MOD_EDGE_TO_LAST_PIPE = 10 * u.cm

NOM_DIAM_FLOC_RM_RESTRAINER = 0.5 * u.inch

###Height that the drain stub extends above the top of the flocculator wall
LENGTH_FLOC_DRAIN_STUB_EXT = 20 * u.cm

SPACE_FLOC_MOD_PIPE_TO_EDGE = 10 * u.cm

THICKNESS_FLOC_BAFFLE = 2 * u.mm



###Sedimentation tank
##General

VEL_SED_UP_BOD = 1 * u.mm/u.s

##Plate settler
VEL_SED_CONC_BOD = 0.12 * u.mm/u.s  # capture velocity

ANGLE_SED_PLATE = 60 * u.deg

SPACE_SED_PLATE = 2.5*u.cm

N_SED_MODULE_PLATES_MIN = 8

# This is moved to template because THICKNESS_SED_PLATE is in materials.yaml
# DIST_CENTER_SED_PLATE = SPACE_SED_PLATE + THICKNESS_SED_PLATE

# Bottom of channel
ANGLE_SED_SLOPE = 50 * u.deg

##This slope needs to be verified for functionality in the field.
# A steeper slope may be required in the floc hopper.
ANGLE_SED_HOPPER_SLOPE = 45 * u.deg

HEIGHT_WATER_SED_EST = 2 * u.m

SED_GATE_VALVE_URL = "https://confluence.cornell.edu/download/attachments/173604905/Sed-Scaled-Gate-Valve-Threaded.dwg"

SUPPORT_BOLT_URL = "https://confluence.cornell.edu/download/attachments/173604905/PlateSettlerSupportBolt.dwg"

##Inlet channel
HEADLOSS_SED_WEIR_MAX = 5 * u.cm

##Height of the inlet channel overflow weir above the normal water level
# in the inlet channel so that the far side of the overflow weir does not
# fill with water under normal operating conditions. This means the water
# level in the inlet channel will increase when the inlet overflow weir
# is in use.
HEIGHT_SED_INLET_WEIR_FREE_BOARD = 2 * u.cm

THICKNESS_SED_WEIR = 5*u.cm

HL_SED_INLET_MAX = 1 * u.cm

# ratio of the height to the width of the sedimentation tank inlet channel.
RATIO_HW_SED_INLET = 0.95

##Exit launder
##Target headloss through the launder orifices
HEADLOSS_SED_LAUNDER_BOD = 4 * u.cm

##Acceptable ratio of min to max flow through the launder orifices
RATIO_FLOW_LAUNDER_ORIFICES = 0.80

##Center to center spacing of orifices in the launder
DIST_CENTER_SED_LAUNDER_EST = 10 * u.cm

##The additional length needed in the launder cap pipe that is to be
# inserted into the launder coupling
LENGTH_SED_LAUNDER_CAP_EXCESS = 3 * u.cm

##Space between the top of the plate settlers and the bottom of the
# launder pipe
HEIGHT_LAMELLA_TO_LAUNDER = 5 * u.cm

##The additional length needed in the launder cap pipe that is to be
# inserted into the launder coupling

##Diameter of the pipe used to hold the plate settlers together
NOM_DIAMETER_SED_MOD = 0.5 * u.inch

##Diameter of the pipe used to create spacers. The spacers slide over the
# 1/2" pipe and are between the plates
NOM_DIAMETER_SED_MOD_SPACER = 0.75 * u.inch

SDR_SED_MOD_SPACER = 17

##This is the vertical thickness of the lip where the lamella support sits. mrf222
THICKNESS_SED_LAMELLA_LEDGE = 8 * u.cm

SPACE_SED_LAMELLA_PIPE_TO_EDGE = 5 * u.cm

##Approximate x-dimension spacing between cross pipes in the plate settler
# support frame.
DIST_CENTER_SED_PLATE_FRAME_CROSS_EST = 0.8 * u.m

##Estimated plate length used to get an initial estimate of sedimentation
# tank active length.
LENGTH_SED_PLATE_EST = 60 * u.cm

##Pipe size of the support frame that holds up the plate settler modules
NOM_DIAMETER_SED_PLATE_FRAME = 1.5 * u.inch

##Floc weir

#Vertical distance from the top of the floc weir to the bottom of the pipe
# frame that holds up the plate settler modules
HEIGHT_FLOC_WEIR_TO_PLATE_FRAME = 10 * u.cm

##Minimum length (X dimension) of the floc hopper
LENGTH_SED_HOPPER_MIN = 50 * u.cm

##Inlet manifold
##Max energy dissipation rate in the sed diffuser outletS
ENERGY_DIS_SED_INT_MAX = 150 * u.mW/u.kg

##Ratio of min to max flow through the inlet manifold diffusers
RATIO_FLOW_SED_INLET = 0.8

ND_SED_MANIFOLD_MAX = 8 * u.inch

SDR_SED_MANIFOLD = 41  # SDR of pipe for sed tank inlet manifold

##This is the minimum distance between the inlet manifold and the slope
# of the sed tank.
SPACE_SED_INLET_MAN_SLOPE = 10 * u.cm

##Length of exposed manifold stub coming out of the floc weir to which the
# free portion of the inlet manifold is attached with a flexible coupling.
LENGTH_SED_MAN_CONNECTION_STUB = 4 * u.cm

##Space between the end of the manifold pipe and the edge of the first
# diffuser's hole, or the first manifold orifice.

LENGTH_SED_MANIFOLD_FIRST_DIFFUSER_GAP = 3 * u.cm

##Vertical distance from the edge of the jet reverser half-pipe to the tip
# of the inlet manifold diffusers
HEIGHT_JET_REVERSER_TO_DIFFUSERS = 3 * u.cm

##Gap between the end of the inlet manifold pipe and the end wall of the
# tank to be able to install the pipe
LENGTH_SED_MANIFOLD_PIPE_FROM_TANK_END = 2  *u.cm

LENGTH_SED_WALL_TO_DIFFUSER_GAP_MIN = 3 * u.cm

##Diameter of the holes drilled in the manifold so that the molded 1"
# diffuser pipes can fit tightly in place (normal OD of a 1" pipe is
# close to 1-5/16")
DIAM_SED_MANIFOLD_PORT = 1.25 * u.inch

ND_JET_REVERSER = 3 * u.inch  # nominal diameter of pipe used for jet reverser in bottom of set tank

SDR_REVERSER = 26  # SDR of jet reverser pipe

## Diffuser geometry
SDR_DIFFUSER = 26  # SDR of diffuser pipe

ND_DIFFUSER_PIPE = 4 * u.cm  # nominal diameter of pipe used to make diffusers

AREA_PVC_DIFFUSER = (np.pi/4) * ((pipe.OD(ND_DIFFUSER_PIPE)**2)
                                 - (pipe.ID_SDR(ND_DIFFUSER_PIPE, SDR_DIFFUSER))**2)

RATIO_PVC_STRETCH = 1.2  # stretch factor applied to the diffuser PVC pipes as they are heated and molded

T_DIFFUSER = ((pipe.OD(ND_DIFFUSER_PIPE) -
                        pipe.ID_SDR(ND_DIFFUSER_PIPE, SDR_DIFFUSER))
                              / (2 * RATIO_PVC_STRETCH))

W_DIFFUSER_INNER = 0.3175 * u.cm  # opening width of diffusers

# Calculating using a minor loss equation with K = 1
V_SED_DIFFUSER_MAX = np.sqrt(2 * GRAVITY * HL_SED_INLET_MAX).to(u.mm/u.s)

L_DIFFUSER = 15 * u.cm  # vertical length of diffuser

B_DIFFUSER = 5 * u.cm  # center to center spacing beteen diffusers

HEADLOSS_SED_DIFFUSER = 0.001 * u.m # Headloss through the diffusers to ensure uniform flow between sed tanks

##Outlet to filter
#If the plant has two trains, the current design shows the exit channel
# continuing from one set of sed tanks into the filter inlet channel.
#The execution of this extended channel involves a few calculations.
HEADLOSS_SED_TO_FILTER_PIPE_MAX = 10 * u.cm
#==============================================================================
# if EN_DOUBLE_TRAIN == 1:
#     K_SED_EXIT = 1
# else:
#     K_SED_EXIT = 0
#
#
# if EN_DOUBLE_TRAIN == 1:
#    HEIGHT_EXIT_FREE = 5 * u.cm
# else:
#    HEIGHT_EXIT_FREE = 0 * u.cm
#==============================================================================

##added 12/5/16 by mrf222 ensures weir does not overtop backwards if
# filter weir is too high
HEIGHT_SED_WEIR_FREE_BOARD = 5 * u.cm



##Stacked rapid sand filter
####Construction and Design Inputs

#Design guidelines say 11 mm/s. The success of lab-scale backwashing at
# 10 mm/s suggests that this is a reasonable and conservative value
VEL_FILTER_Bw_ = 11 * u.mm/u.s

N_FILTER_LAYER = 6

VEL_FILTER_LAYER = 1.833 * u.mm/u.s ##VEL_FIBER_DIST_CENTER_/N_FIBER_LAYER

##Minimum thickness of each filter layer (can be increased to accomodate
# larger pipe diameters in the bottom layer)
HEIGHT_FILTER_LAYER_MIN = 20 * u.cm

##center to center distance for slotted pipes
DIST_CENTER_FILTER_MANIFOLD_BRANCH = 10 * u.cm

##How far the branch extends into the trunk line
LENGTH_FILTER_MAN_BRANCH_EXT = 2 * u.cm

##The time to drain the filter box of the water above the fluidized bed
TIME_FIBER_BACKWASH_INITIATION_BOD = 3 * u.min

##Mickey suggested this value based on lab experience. This was moved to
# Expert Inputs 12/4/16 by mrf222 as a result of feedback from Monroe and
# Skyler. In the Moroceli plant, the Fi Entrance box was overflowing
# before filtration backwash. The HL of a dirty filter has therefore been
# increased from 40 to 60 cm.
HEADLOSS_FILTER_DIRTY = 60 * u.cm

##This is the extra head we are going to provide on top of steady state
# backwash head loss to ensure that we can fluidize the bed to initiate
# backwash.
HEADLOSS_FIBER_BACKWASH_STEADY_FLOW = 20 * u.cm

##Maximum acceptable head loss through the siphon at steady state; used to
# calculate a diameter
HEADLOSS_FILTER_SIPHON_MAX = 35 * u.cm

##Diameter of sand drain pipe
NOM_DIAMETER_FILTER_SAND_OUTLET = 2 * u.inch

##Height of the barrier between the exit box and distribution box.
HEIGHT_FILTER_DIST_BARRIER = 10 * u.cm

##Length that the siphon pipe extends up into the plant drain channel.
#Being able to shorten the stub from which the siphon discharges into the
# main plant drain channel allows for some flexibility in the hydraulic design.
LENGTH_FILTER_SIPHON_CHANNEL_STUB_MIN = 20 * u.cm

HEADLOSS_FILTER_ENTRANCE_PIPE_MAX = 10 * u.cm

NOM_DIAMETER_FILTER_TRUNK_MAX = 6 * u.inch

NOM_DIAMETER_FILTER_BACK_WASH_SIPHON_MAX = 8 * u.inch

##Purge valves on the trunk lines are angled downwards so that sediment is
# cleared more effectively. This angle allows the tees to fit on top of one
# another at the filter wall.
ANGLE_FILTER_TRUNK_VALVES = 25 * u.deg

##Purge valves on the trunk lines are angled downwards so that sediment is
# cleared more effectively. This angle allows the tees to fit on top of one
# another at the filter wall.
THICKNESS_FILTER_WEIR = 5 * u.cm

SPACE_FILTER_BRANCH_TO_WALL = 5 * u.cm

FILTER_GATE_VALUE_URL = "https://confluence.cornell.edu/download/attachments/173604905/Fi-Scaled-Gate-Valve-Threaded.dwg"
FILTER_BALL_VALVE_URL = "https://confluence.cornell.edu/download/attachments/173604905/FiMetalBallValve.dwg"

HEIGHT_FILTER_WALL_TO_PLANT_FLOOR_MIN = 10 * u.cm

HEADLOSS_FILTER_INLET_WEIR_MAX = 5 * u.cm

##Dimensions get too small for construction below a certain flow rate
FLOW_FILTER_MIN = 8 * u.L/u.s

LENGTH_FILTER_MAN_FEMCO_COUPLING = 6 * u.cm

##Nominal diameter of the spacer tees in the four corners of the filter
# manifold assembly.
NOM_DIAMETER_FILTER_MAN_WING_SPACER = 2 * u.inch

##Length of the vertical pipe segment following the valve on the filter
# sand drain. This stub can be capped to allow the sand in the valve to
# settle, so that the valve can be closed without damage from fluidized sand.
LENGTH_FILTER_SAND_OUTLET_PIPE = 20 * u.cm




#######Elevation Safety Margins

##Minimum depth in the entrance box during backwash such that there is
# standing water over the inlet.
HEIGHT_FILTER_BACKWASH_NO_SUCK_AIR = 20 * u.cm

##Minimum water depth over the orifices in the siphon manifold so that air
# is not entrained.
HEIGHT_FILTER_SIPHON_NO_SUCK_AIR = 10 * u.cm

HEIGHT_FILTER_FLUIDIZED_BED_TO_SIPHON = 20 * u.cm

HEIGHT_FILTER_FORWARD_NO_SUCK_AIR = 10 * u.cm

HEIGHT_FILTER_WEIR_FREEFALL = 3 * u.cm

HEIGHT_FILTER_AIR_REMOVAL_BLOCK_SUBMERGED = 5 * u.cm

HEIGHT_FILTER_BYPASS_SAFETY = 10 * u.cm

HEIGHT_DRAIN_OUTLET_SAFETY = 10 * u.cm

HEIGHT_FILTER_OVERFLOW_WEIR_FREEFALL = 10 * u.cm



#######Plant drain channel

###Space beyond the entrance tank in the plant drain channel where the
# drop pipes from the CDC lever arm can come down and be connected with
# the chlorine and coagulant dosing points.
LENGTH_CHEM_LEVER_ARM_SPACE = 75 * u.cm


###Operator access

##combine walkway assumptions!
WIDTH_MP_WALKWAY_MIN = 1 * u.m

##Width of the walkway above the main plant drain channel
WIDTH_DC_WALKWAY = 1.2 * u.m

##Width of the floor space between the flocculator and the rapid mix pipe
# floor cutout next to the entrance tank.
WIDTH_ET_WALKWAY = 1 * u.m

##for high flow, double train situations
W_TRAIN_WALKWAY = 1.5 * u.m

W_BASEMENT_STAIRS = 0.9 * u.m

W_ENTRANCE_STAIRS = 1.2 * u.m


##Minor loss coefficients
##Individual K Values

##90 deg elbow
K_MINOR_EL90 = 0.9

K_MINOR_EL45 = 0.45
##The loss coefficient for the channel transition in a 90 degree turn
K_MINOR_90 = 0.4

K_MINOR_ANGLE_VALVE = 4.3

K_MINOR_GLOBE_VALVE = 10

K_MINOR_GATE_VALVE = 0.39

K_MINOR_CHECK_VALVE_CONV = 4

K_MINOR_CHECK_VALVE_BALL = 4.5

##headloss coefficient of jet
K_MINOR_EXP = 1

K_MINOR_TEE_FLOW_RUN = 0.6

K_MINOR_TEE_FLOW_BR = 1.8

K_MINOR_PIPE_ENTRANCE = 0.5

K_MINOR_PIPE_EXIT = 1

K_MINOR_RM_GATE_VIN = 25



"""Module containing global `pint` unit registry.

The `pint` module supports arithmetic involving *physical quantities*
each of which has a magnitude and units, for example 1 cm or 3 kg.
The units of a quantity come from a `pint` *unit registry*, and it
appears that `pint` supports arithmetic operations only on quantities
whose units come from the same unit registry (an attempt to perform
an operation on quantities whose units come from different unit
registries raises an exception). This module contains a single global
unit registry `unit_registry` that can be used by any number of other
modules.

"""

import os
import pint
import pandas as pd

unit_registry = pint.UnitRegistry(system='mks', autoconvert_offset_to_baseunit=True)

# default formatting includes 4 significant digits. This can be overridden on a per-print basis
#  with print('{:.3f}'.format(3*ureg.m /9))
unit_registry.default_format = '.4g'
pd.options.display.float_format = '{:,.4g}'.format


def set_sig_fig(n: int):
    """Set the default number of significant figures used to print pint, pandas and numpy values
    quantities. Defaults to 4.

    Args:
        n: number of significant figures to display

    Example:
        import aguaclara
        from aguaclara.units import unit_registry as u
        h=2.5532532522352543*u.m
        e = 25532532522352543*u.m
        print('h before sigfig adjustment: ',h)
        print('e before sigfig adjustment: ',e)
        aguaclara.units.set_sig_fig(10)
        print('h after sigfig adjustment: ',h)
        print('e after sigfig adjustment: ',e)

        h before sigfig adjustment:  2.553 meter
        e before sigfig adjustment:  2.553e+16 meter
        h after sigfig adjustment:  2.553253252 meter
        e after sigfig adjustment:  2.553253252e+16 meter
    """
    unit_registry.default_format = '.' + str(n) + 'g'
    pd.options.display.float_format = ('{:,.' + str(n) + '}').format

unit_registry.load_definitions(os.path.join(os.path.dirname(__file__), "data/unit_definitions.txt"))




"""This file contains all the functions needed to design the linear flow
orifice meter (LFOM) for an AguaClara plant.

"""

#Here we import packages that we will need for this notebook. You can find out about these packages in the Help menu.
from aguaclara.play import*

#primary outputs from this file are
#Nominal diameter nom_diam_lfom_pipe(FLOW,HL_LFOM,con.RATIO_LFOM_SAFETY)
#number of rows n_rows(FLOW,HL_LFOM)
#orifice diameter orifice_diameter(FLOW,HL_LFOM,drill_series_uom)
#number of orifices in each row n_lfom_orifices(FLOW,HL_LFOM,drill_series_uom)
#height of the center of each row height_lfom_orifices(FLOW,HL_LFOM,drill_series_uom)

# output is width per flow rate.
@u.wraps(u.s/(u.m**2), [u.m,u.m], False)
def width_stout(HL_LFOM,z):
    return (2/((2*pc.gravity*z)**(1/2)*con.RATIO_VC_ORIFICE*np.pi*HL_LFOM)).magnitude


@u.wraps(None, [u.m**3/u.s,u.m], False)
def n_lfom_rows(FLOW,HL_LFOM):
    """This equation states that the open area corresponding to one row can be
    set equal to two orifices of diameter=row height. If there are more than
    two orifices per row at the top of the LFOM then there are more orifices
    than are convenient to drill and more than necessary for good accuracy.
    Thus this relationship can be used to increase the spacing between the
    rows and thus increase the diameter of the orifices. This spacing function
    also sets the lower depth on the high flow rate LFOM with no accurate
    flows below a depth equal to the first row height.

    But it might be better to always set then number of rows to 10.
    The challenge is to figure out a reasonable system of constraints that
    reliably returns a valid solution.
    """
    N_estimated = (HL_LFOM*np.pi/(2*width_stout(HL_LFOM,HL_LFOM)*FLOW))
    variablerow = min(10,max(4,math.trunc(N_estimated.magnitude)))
    # Forcing the LFOM to either have 4 or 8 rows, for design purposes
    # If the hydraulic calculation finds that there should be 4 rows, then there
    # will be 4 rows. If anything other besides 4 rows is found, then assign 8
    # rows.
    # This can be improved in the future.
    if variablerow == 4:
        variablerow = 4
    else:
        variablerow = 8
    return variablerow

@u.wraps(u.m, [u.m**3/u.s, u.m], False)
def dist_center_lfom_rows(FLOW,HL_LFOM):
    return HL_LFOM/n_lfom_rows(FLOW,HL_LFOM)

@u.wraps(u.m/u.s, [u.m], False)
def vel_lfom_pipe_critical(HL_LFOM):
    """The average vertical velocity of the water inside the LFOM pipe
    at the very bottom of the bottom row of orifices
    The speed of falling water is 0.841 m/s for all linear flow orifice meters
    of height 20 cm, independent of total plant flow rate."""
    return 4/(3*math.pi)*(2*pc.gravity.magnitude*HL_LFOM)**(1/2)

@u.wraps(u.m**2, [u.m**3/u.s, u.m], False)
def area_lfom_pipe_min(FLOW, HL_LFOM):
    return (con.RATIO_LFOM_SAFETY*FLOW/vel_lfom_pipe_critical(HL_LFOM).magnitude)

@u.wraps(u.inch, [u.m**3/u.s, u.m], False)
def nom_diam_lfom_pipe(FLOW,HL_LFOM):
    ID = pc.diam_circle(area_lfom_pipe_min(FLOW, HL_LFOM))
    return pipe.ND_SDR_available(ID, mat.SDR_LFOM).magnitude

@u.wraps(u.m**2, [u.m**3/u.s, u.m], False)
def area_lfom_orifices_top(FLOW,HL_LFOM):
    """Estimate the orifice area corresponding to the top row of orifices.
    Another solution method is to use integration to solve this problem.
    Here we use the width of the stout weir in the center of the top row
    to estimate the area of the top orifice
    """
    return ((FLOW*width_stout(HL_LFOM*u.m,HL_LFOM*u.m-0.5*dist_center_lfom_rows(FLOW,HL_LFOM)).magnitude *
        dist_center_lfom_rows(FLOW,HL_LFOM).magnitude))

@u.wraps(u.m, [u.m**3/u.s, u.m], False)
def d_lfom_orifices_max(FLOW, HL_LFOM):
    return (pc.diam_circle(area_lfom_orifices_top(FLOW,HL_LFOM).magnitude).magnitude)

@u.wraps(u.m, [u.m**3/u.s, u.m, u.inch], False)
def orifice_diameter(FLOW,HL_LFOM,drill_bits):
    maxdrill = (min((dist_center_lfom_rows(FLOW,HL_LFOM).magnitude),(d_lfom_orifices_max(FLOW,HL_LFOM).magnitude)))
    return ut.floor_nearest(maxdrill,drill_bits)

@u.wraps(u.m**2, [u.m**3/u.s, u.m, u.inch], False)
def drillbit_area(FLOW,HL_LFOM,drill_bits):
    return pc.area_circle(orifice_diameter(FLOW,HL_LFOM,drill_bits)).magnitude

@u.wraps(None, [u.m**3/u.s, u.m, u.inch], False)
def n_lfom_orifices_per_row_max(FLOW,HL_LFOM,drill_bits):
    """A bound on the number of orifices allowed in each row.
    The distance between consecutive orifices must be enough to retain
    structural integrity of the pipe.
    """
    return math.floor(math.pi*(pipe.ID_SDR(
        nom_diam_lfom_pipe(FLOW, HL_LFOM), mat.SDR_LFOM).magnitude)
        / (orifice_diameter(FLOW, HL_LFOM, drill_bits).magnitude +
            opt.S_LFOM_ORIFICE.magnitude))

@u.wraps(u.m**3/u.s, [u.m**3/u.s, u.m], False)
def flow_ramp(FLOW,HL_LFOM):
    n_rows = n_lfom_rows(FLOW,HL_LFOM)
    return np.linspace(FLOW/n_rows,FLOW,n_rows)

@u.wraps(u.m, [u.m**3/u.s, u.m, u.inch], False)
def height_lfom_orifices(FLOW,HL_LFOM,drill_bits):
    """Calculates the height of the center of each row of orifices.
    The bottom of the bottom row orifices is at the zero elevation
    point of the LFOM so that the flow goes to zero when the water height
    is at zero.
    """
    return (np.arange((orifice_diameter(FLOW,HL_LFOM,drill_bits)*0.5),
                      HL_LFOM,
                      (dist_center_lfom_rows(FLOW,HL_LFOM))))

@u.wraps(u.m**3/u.s, [u.m**3/u.s, u.m, u.inch, None, None], False)
def flow_lfom_actual(FLOW,HL_LFOM,drill_bits,Row_Index_Submerged,N_LFOM_Orifices):
    """Calculates the flow for a given number of submerged rows of orifices
    harray is the distance from the water level to the center of the orifices
    when the water is at the max level
    """
    D_LFOM_Orifices=orifice_diameter(FLOW, HL_LFOM, drill_bits).magnitude
    row_height=dist_center_lfom_rows(FLOW, HL_LFOM).magnitude
    harray = (np.linspace(row_height, HL_LFOM, n_lfom_rows(FLOW, HL_LFOM))) - 0.5 * D_LFOM_Orifices
    FLOW_new = 0
    for i in range(Row_Index_Submerged+1):
        FLOW_new = FLOW_new + (N_LFOM_Orifices[i]*(
            pc.flow_orifice_vert(D_LFOM_Orifices, harray[Row_Index_Submerged-i],
                con.RATIO_VC_ORIFICE).magnitude))
    return FLOW_new


#Calculate number of orifices at each level given a diameter
@u.wraps(None, [u.m**3/u.s, u.m, u.inch], False)
def n_lfom_orifices(FLOW,HL_LFOM,drill_bits):
    FLOW_ramp_local = flow_ramp(FLOW,HL_LFOM).magnitude
    n_orifices_max =n_lfom_orifices_per_row_max(FLOW,HL_LFOM,drill_bits)
    n_rows = (n_lfom_rows(FLOW,HL_LFOM))
    D_LFOM_Orifices = orifice_diameter(FLOW,HL_LFOM,drill_bits).magnitude
    # H is distance from the elevation between two rows of orifices down to the center of the orifices
    H=dist_center_lfom_rows(FLOW,HL_LFOM).magnitude-D_LFOM_Orifices*0.5
    n=[]
    for i in range(n_rows):
        #place zero in the row that we are going to calculate the required number of orifices
        n=np.append(n,0)
        #calculate the ideal number of orifices at the current row without constraining to an integer
        n_orifices_real=((FLOW_ramp_local[i]-flow_lfom_actual(FLOW,HL_LFOM,drill_bits,i,n).magnitude)/
                                  pc.flow_orifice_vert(D_LFOM_Orifices,H,con.RATIO_VC_ORIFICE)).magnitude
        #constrain number of orifices to be less than the max per row and greater or equal to 0
        n[i]=min((max(0,round(n_orifices_real))),n_orifices_max)
    return n

#This function takes the output of n_lfom_orifices and converts it to a list with 8
#entries that corresponds to the 8 possible rows. This is necessary to make the lfom
# easier to construct in Fusion using patterns
@u.wraps(None, [u.m**3/u.s, u.m, u.inch, None], False)
def n_lfom_orifices_fusion(FLOW,HL_LFOM,drill_bits,num_rows):
    num_orifices_per_row = n_lfom_orifices(FLOW, HL_LFOM, drill_bits)
    num_orifices_final = np.zeros(8)
    centerline = np.zeros(8)
    center = True
    for i in range(8):
        if i % 2 == 1 and num_rows == 4:
            centerline[i] = int(center)
        elif num_rows == 4:
            num_orifices_final[i] = num_orifices_per_row[i/2]
            centerline[i] = int(center)
            center = not center
        else:
            num_orifices_final[i] = num_orifices_per_row[i]
            centerline[i] = int(center)
            center = not center

    return num_orifices_final, centerline



#This function calculates the error of the design based on the differences between the predicted flow rate
#and the actual flow rate through the LFOM.
@u.wraps(u.m**3/u.s, [u.m**3/u.s, u.m, u.inch], False)
def flow_lfom_error(FLOW,HL_LFOM,drill_bits):
    N_lfom_orifices=n_lfom_orifices(FLOW,HL_LFOM,drill_bits,mat.SDR_LFOM)
    FLOW_lfom_error=[]
    for j in range (len(N_lfom_orifices)-1):
        FLOW_lfom_error.append((flow_lfom_actual(
            FLOW, HL_LFOM, drill_bits, j, N_lfom_orifices).magnitude-flow_ramp(
            FLOW, HL_LFOM)[j].magnitude)/FLOW)
    return FLOW_lfom_error


@u.wraps(u.m**3/u.s, [u.m**3/u.s, u.m, u.m], False)
def flow_lfom_ideal(FLOW, HL_LFOM, H):
    flow_lfom_ideal=(FLOW*H)/HL_LFOM
    return flow_lfom_ideal

@u.wraps(u.m**3/u.s, [u.m**3/u.s, u.m, u.inch, u.m], False)
def flow_lfom(FLOW,HL_LFOM,drill_bits,H):
    D_lfom_orifices=orifice_diameter(FLOW,HL_LFOM,drill_bits).magnitude
    H_submerged=np.arange(H-0.5*D_lfom_orifices, HL_LFOM, H-dist_center_lfom_rows(FLOW,HL_LFOM).magnitude,dtype=object)
    N_lfom_orifices=n_lfom_orifices(FLOW,HL_LFOM,drill_bits,mat.SDR_LFOM)
    flow=[]
    for i in range(len(H_submerged)):
        flow.append(pc.flow_orifice_vert(D_lfom_orifices,H_submerged[i],con.RATIO_VC_ORIFICE)*N_lfom_orifices[i])
    return sum(flow)



"""This file contains all the functions needed to design a flocculator for
an AguaClara plant.

"""
from aguaclara.play import*

# expansion minor loss coefficient for 180 degree bend
K_e = (1 / con.RATIO_VC_ORIFICE**2 - 1)**2

@u.wraps(1/u.s, [u.m, None, u.degK], False)
def G_avg(hl, Gt, T):
    """Return the average velocity gradient of a flocculator given head
    loss, collision potential and temperature.

    Parameters
    ----------
    hl: float
        Headloss through the flocculator

    Gt: float
        Target collision potential

    T: float
        Design temperature

    Returns
    -------
    ?

    Examples
    --------
    >>> from aguaclara.play import*
    >>>G_avg(40*u.cm, 37000, 25*u.degC)
    118.715 1/second
    """
    G = (pc.gravity.magnitude * hl) / (Gt * pc.viscosity_kinematic(T).magnitude)
    return G

@u.wraps(u.m**3, [u.m**3/u.s, u.m, None, u.degK], False)
def vol_floc(q_plant, hl, Gt, T):
    """Return the total volume of the flocculator using plant flow rate, head
    loss, collision potential and temperature.

    Uses an estimation of flocculator residence time (ignoring the decrease
    in water depth caused by head loss in the flocculator.) Volume does not take
    into account the extra volume that the flocculator will have due to changing
    water level caused by head loss.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    hl: float
        Headloss through the flocculator

    Gt: float
        Target collision potential

    T: float
        Design temperature

    Returns
    -------
    ?

    Examples
    --------
    >>> from aguaclara.play import*
    >>>vol_floc(20*u.L/u.s, 40*u.cm, 37000, 25*u.degC)
    6.233 meter3
    """
    vol = (Gt / G_avg(hl, Gt, T).magnitude)*q_plant
    return vol

@u.wraps(u.cm, [u.m**3/u.s, u.m, None, u.degK, u.m], False)
def width_HS_min(q_plant, hl, Gt, T, depth_end):
    """Return the minimum channel width required to achieve H/S > 3.

    The channel can be wider than this, but this is the absolute minimum width
    for a channel. The minimum width occurs when there is only one expansion per
    baffle and thus the distance between expansions is the same as the depth of
    water at the end of the flocculator.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    hl: float
        Headloss through the flocculator

    Gt: float
        Target collision potential

    T: float
        Design temperature

    depth_end: float
        The depth of water at the end of the flocculator

    Returns
    -------
    float
        The minimum channel width required to achieve H/S > 3

    Examples
    --------
    >>> from aguaclara.play import*
    >>> width_HS_min(20*u.L/u.s, 40*u.cm, 37000, 25*u.degC, 2*u.m)
    0.1074 centimeter
    """
    nu = pc.viscosity_kinematic(T).magnitude

    w = con.RATIO_HS_MIN*((K_e/(2 * depth_end * (G_avg(hl, Gt, T).magnitude**2)
    * nu))**(1/3))*q_plant/depth_end
    return w



@u.wraps(u.cm, [u.m**3/u.s, u.m, None, u.degK, u.m], False)
def width_floc_min(q_plant, hl, Gt, T, depth_end):
    """Return the minimum channel width required.

    This takes the maximum of the minimum required to achieve H/S > 3 and the
    minimum required for constructability based on the width of the human hip.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    hl: float
        Headloss through the flocculator

    Gt: float
        Target collision potential

    T: float
        Design temperature

    depth_end: float
        The depth of water at the end of the flocculator

    Returns
    -------
    float
        The minimum channel width required to achieve H/S > 3

    Examples
    --------
    >>> from aguaclara.play import*
    >>> width_floc_min(20*u.L/u.s, 40*u.cm, 37000, 25*u.degC, 2*u.m)
    45 centimeter
    """
    return max(width_HS_min(q_plant, hl, Gt, T, depth_end).magnitude, con.FLOC_WIDTH_MIN_CONST.magnitude)

@u.wraps(None, [u.m**3/u.s, u.m, None, u.degK, u.m, u.m], False)
def num_channel(q_plant, hl, Gt, T, W_tot, depth_end):
    """Return the number of channels in the entrance tank/flocculator (ETF).

    This takes the total width of the flocculator and divides it by the minimum
    channel width. A floor function is used to ensure that there are an even
    number of channels.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    hl: float
        Headloss through the flocculator

    Gt: float
        Target collision potential

    T: float
        Design temperature

    W_tot: float
        Total width

    depth_end: float
        The depth of water at the end of the flocculator

    Returns
    -------
    ?

    Examples
    --------
    >>> from aguaclara.play import*
    >>> num_channel(20*u.L/u.s, 40*u.cm, 37000, 25*u.degC, 20*u.m, 2*u.m)
    2
    """
    num = W_tot/(width_floc_min(q_plant, hl, Gt, T, depth_end).magnitude)
    # floor function with step size 2
    num = np.floor(num/2)*2
    return int(max(num, 2))


@u.wraps(u.m**2, [u.m**3/u.s, u.m, None, u.degK, u.m], False)
def area_ent_tank(q_plant, hl, Gt, T, depth_end):
    """Return the planview area of the entrance tank given plant flow rate,
    headloss, target collision potential, design temperature, and depth of
    water at the end of the flocculator.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    hl: float
        Headloss through the flocculator

    Gt: float
        Target collision potential

    T: float
        Design temperature

    depth_end: float
        The depth of water at the end of the flocculator

    Returns
    -------
    float
        The planview area of the entrance tank

    Examples
    --------
    >>> from aguaclara.play import*
    >>> area_ent_tank(20*u.L/u.s, 40*u.cm, 37000, 25*u.degC, 2*u.m)
    1 meter ** 2
    """
    # guess the planview area before starting iteration
    A_new = 1*u.m**2
    A_ratio = 0

    while (A_ratio) > 1.01 and (A_ET_PV/A_new) < 0.99:
        A_ET_PV = A_new

        vol_floc = vol_floc(q_plant, hl, Gt, T)
        A_floc_PV = vol_floc/(depth_end + hl/2)
        A_ETF_PV = A_ET_PV + A_floc_PV

        W_min = width_floc_min(q_plant, hl, Gt, T, depth_end)

        W_tot = A_ETF_PV/opt.L_sed

        num_chan = num_channel(q_plant, hl, Gt, T, W_tot)
        W_chan = W_tot/num_chan

        A_new = opt.L_ET_max*W_chan

        A_ratio = A_new/A_ET_PV

    return A_new.to(u.m**2).magnitude

### Baffle calculations
@u.wraps(u.m, [u.m**3/u.s, u.m, None, u.degK, u.m], False)
def exp_dist_max(q_plant, hl, Gt, T, W_chan):
    """"Return the maximum distance between expansions for the largest
    allowable H/S ratio.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    hl: float
        Headloss through the flocculator

    Gt: float
        Target collision potential

    T: float
        Design temperature

    W_chan: float
        Channel width

    Returns
    -------
    ?

    Examples
    --------
    >>> from aguaclara.play import*
    >>> exp_dist_max(20*u.L/u.s, 40*u.cm, 37000, 25*u.degC, 2*u.m)
    0.375 meter
    """
    g_avg = G_avg(hl, Gt, T).magnitude
    nu = pc.viscosity_kinematic(T).magnitude
    term1 = (K_e/(2 * (g_avg**2) * nu))**(1/4)
    term2 = (con.RATIO_HS_MAX*q_plant/W_chan)**(3/4)
    exp_dist_max = term1*term2
    return exp_dist_max

@u.wraps(None, [u.m**3/u.s, u.m, None, u.degK, u.m], False)
def num_expansions(q_plant, hl, Gt, T, depth_end):
    """"Return the minimum number of expansions per baffle space.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    hl: float
        Headloss through the flocculator

    Gt: float
        Target collision potential

    T: float
        Design temperature

    depth_end: float
        The depth of water at the end of the flocculator

    Returns
    -------
    ?

    Examples
    --------
    >>> from aguaclara.play import*
    ???
    """
    return int(np.ceil(depth_end/(exp_dist_max(q_plant, hl, Gt, T)).magnitude))

@u.wraps(u.m, [u.m**3/u.s, u.m, None, u.degK, u.m], False)
def height_exp(q_plant, hl, Gt, T, depth_end):
    """Return the actual distance between expansions given the integer
    requirement for the number of expansions per flocculator depth.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    hl: float
        Headloss through the flocculator

    Gt: float
        Target collision potential

    T: float
        Design temperature

    depth_end: float
        The depth of water at the end of the flocculator

    Returns
    -------
    ?

    Examples
    --------
    >>> from aguaclara.play import*
    ???
    """
    return depth_end/num_expansions(q_plant, hl, Gt, T)

@u.wraps(u.m, [u.m**3/u.s, u.m, None, u.degK, u.m], False)
def baffle_spacing(q_plant, hl, Gt, T, W_chan):
    """Return the spacing between baffles based on the target velocity gradient

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    hl: float
        Headloss through the flocculator

    Gt: float
        Target collision potential

    T: float
        Design temperature

    W_chan: float
        Channel width

    Returns
    -------
    ?

    Examples
    --------
    >>> from aguaclara.play import*
    >>> baffle_spacing(20*u.L/u.s, 40*u.cm, 37000, 25*u.degC, 2*u.m)
    0.063 meter
    ."""
    g_avg = G_avg(hl, Gt, T).magnitude
    nu = pc.viscosity_kinematic(T).magnitude
    term1 = (K_e/(2 * exp_dist_max(q_plant, hl, Gt, T, W_chan).magnitude * (g_avg**2) * nu))**(1/3)
    return term1 * q_plant/W_chan
baffle_spacing(20*u.L/u.s, 40*u.cm, 37000, 25*u.degC, 2*u.m)

@u.wraps(None, [u.m**3/u.s, u.m, None, u.degK, u.m, u.m, u.m], False)
def num_baffles(q_plant, hl, Gt, T, W_chan, L, baffle_thickness):
    """Return the number of baffles that would fit in the channel given the
    channel length and spacing between baffles.

    Parameters
    ----------
    q_plant: float
        Plant flow rate

    hl: float
        Headloss through the flocculator

    Gt: float
        Target collision potential

    T: float
        Design temperature

    W_chan: float
        Channel width

    L: float
        Length

    baffle_thickness: float
        Baffle thickness

    Returns
    -------
    ?

    Examples
    --------
    >>> from aguaclara.play import*
    >>> num_baffles(20*u.L/u.s, 40*u.cm, 37000, 25*u.degC, 2*u.m, 2*u.m, 2*u.m)
    0
    >>> num_baffles(20*u.L/u.s, 20*u.cm, 37000, 25*u.degC, 2*u.m, 2*u.m, 21*u.m)
    -1
    """
    num = round((L / (baffle_spacing(q_plant, hl, Gt, T, W_chan).magnitude + baffle_thickness)))
    # the one is subtracted because the equation for num gives the number of
    # baffle spaces and there is always one less baffle than baffle spaces due
    # to geometry
    return int(num) - 1



import unittest
import aguaclara.core.k_value_of_reductions_utility as k
from aguaclara.core.units import unit_registry as u
from aguaclara.core import pipedatabase as pipe

""" There are still many cases to test."""


class KValuesCalculationTest(unittest.TestCase):

    # Test Reductions
    def test_k_value_reduction_square_turbulent(self):
        self.assertAlmostEqual(k.k_value_reduction(pipe.OD(4), pipe.OD(2), 4 * u.L/u.s), 5.6677039356929662)

    def test_k_value_reduction_laminar(self):
        self.assertAlmostEqual(k.k_value_reduction(pipe.OD(1), pipe.OD(0.5), 0.1 * u.L / u.s), 2.1802730749680945)

    def test_k_value_reduction_from_very_large_pipe_turbulent(self):
        self.assertAlmostEqual(k.k_value_reduction(pipe.OD(400), pipe.OD(4), 4 * u.L / u.s), 105560.31724275621)

    # Test Expansions
    def test_k_value_expansion_into_large_tank(self):
        self.assertAlmostEqual(k.k_value_expansion(pipe.OD(4), pipe.OD(400), 4 * u.L / u.s), 1.0110511331493719)

    def test_k_value_expansion_into_very_large_pipe_laminar(self):
        self.assertAlmostEqual(k.k_value_expansion(pipe.OD(1), pipe.OD(400), 0.1 * u.L / u.s), 1.0216612503304363)

    # Test Orifices

    # Test private functions
    def test_k_value_thick_orifice_high_headloss(self):
        self.assertAlmostEqual(k._k_value_thick_orifice(0.02, 0.002, 0.000002, 2), 1594340.3320537778)

    def test_k_value_thin_orifice_high_headloss(self):
        self.assertAlmostEqual(k._k_value_thin_sharp_orifice(0.02, 0.002, 2), 1594433.5406999998)

    # Test public function
    def test_k_value_thin_orifice_regular_high_headloss(self):
        self.assertAlmostEqual(k.k_value_orifice(0.02 * u.m, 0.002 * u.m, 0 * u.m, 1*u.L/u.s), 1697.9866773221295)

    def test_k_value_super_thick_orifice_high_headloss(self):
        self.assertAlmostEqual(k.k_value_orifice(pipe.OD(6), pipe.OD(4), 60*u.inch, 1 * u.L / u.s), 1.8350488368427034)

    def test_k_value_thin_orifice(self):
        self.assertAlmostEqual(k.k_value_orifice(pipe.OD(6), pipe.OD(4), 0*u.inch, 1 * u.L / u.s), 3.3497584836648246)

    def test_k_value_thick_orifice(self):
        self.assertAlmostEqual(k.k_value_orifice(pipe.OD(6), pipe.OD(4), 1*u.inch, 1 * u.L / u.s),
                         2.9070736824641181)


if __name__ == '__main__':
    unittest.main()



from research.units import unit_registry as u
import numpy as np
import pandas as pd

# pump rotor radius based on minimizing error between predicted and measured
# values
R_pump = 1.62 * u.cm

# empirically derived correction factor due to the fact that larger diameter
# tubing has more loss ue to space smashed by rollers
k_nonlinear = 13

# maximum and minimum rpms for a 100 rpm pump
min_rpm = 3 * u.rev/u.min
max_rpm = 95 * u.rev/u.min


def Q6_roller(ID_tube):
    """This function calculates the volume per revolution of a 6 roller pump
    given the innner diameter (ID) of 3-stop tubing. It was empirically derived
    using the table found at
    http://www.ismatec.com/int_e/pumps/t_mini_s_ms_ca/tubing_msca2.htm

    Parameters
    ----------
    ID_tube : float
        inner diameter of the tube

    Returns
    -------
    float
        flow from the 6 roller pump (mL/rev)

    Examples
    --------
    >>> Q6_roller(2.79*u.mm)
    0.4005495805189351 milliliter/rev
    >>> Q6_roller(1.52*u.mm)
    0.14884596727278446 milliliter/rev
    >>> Q6_roller(0.51*u.mm)
    0.01943899117521222 milliliter/rev

    """
    term1 = (R_pump * 2 * np.pi - k_nonlinear * ID_tube) / u.rev
    term2 = np.pi * (ID_tube ** 2) / 4
    return (term1 * term2).to(u.mL/u.rev)


def ID_colored_tube(color):
    """This function looks up the inner diameter of a tube from the tubing data
    table given the color.

    Parameters
    ----------
    color : string
        color of the tubing to be used

    Returns
    -------
    float
        diameter of the tubing (mm)

    Examples
    --------
    >>> ID_colored_tube("yellow-blue")
    1.52 millimeter
    >>> ID_colored_tube("orange-yellow")
    0.51 millimeter
    >>> ID_colored_tube("purple-white")
    2.79 millimeter

    """
    df = pd.read_csv("/data/tubing_data.txt", delimiter='\t')
    idx = df["Color"] == color
    return df[idx]['Diameter (mm)'].values[0] * u.mm


def C_stock_max(Q_plant, C, tubing_color):
    """This function calculates the maximum stock concentration of a generic
    material given desired concentration in the plant.

    Parameters
    ----------
    Q_plant : float
        flow rate of the plant

    C : float
        desired concentration of the material within the plant

    tubing_color : string
        color of the tubing to be used

    Returns
    -------
    float
        maximum stock concentration (g/L)

    Examples
    --------
    >>> C_stock_max(7*u.mL/u.s, 100*u.NTU, "yellow-blue")
    159.89684125188708 gram/liter
    >>> C_stock_max(7*u.mL/u.s, 2*u.mg/u.L, "orange-yellow")
    14.404039668326215 gram/liter

    """
    ID_tube = ID_colored_tube(tubing_color)
    return (C * Q_plant / (Q6_roller(ID_tube) * min_rpm)).to(u.g/u.L)


def Q_stock_max(Q_plant, C, tubing_color):
    """This function calculates the flow rate of the stock of the desired
    concentration.

    Parameters
    ----------
    Q_plant : float
        flow rate of the plant

    C : float
        desired concentration within the plant

    tubing_color : string
        color of the tubing to be used

    Returns
    -------
    float
        flow rate of the stock (mL/s)

    Examples
    --------
    >>> Q_stock_max(7*u.mL/u.s, 100*u.NTU, "yellow-blue")
    0.007442298363639224 milliliter/second
    >>> Q_stock_max(7*u.mL/u.s, 2*u.mg/u.L, "orange-yellow")
    0.0009719495587606109 milliliter/second

    """
    return (C * Q_plant / C_stock_max(Q_plant, C, tubing_color)).to(u.mL/u.s)


def T_stock(Q_plant, C, tubing_color, V_stock):
    """This function calculates the time after the experiment at which the
    stock container will run out.

    Parameters
    ----------
    Q_plant : float
        flow rate of the plant

    C : float
        desired concentration within the plant

    tubing_color : string
        color of the tubing to be used

    V_stock : float
        volume of the stock container

    Returns
    -------
    float
        time before the stock is depleted (hr)

    Examples
    --------
    >>> T_stock(7*u.mL/u.s, 100*u.NTU, "yellow-blue", 1*u.L)
    37.324192635827984 hour
    >>> T_stock(7*u.mL/u.s, 2*u.mg/u.L, "orange-yellow", 1*u.L)
    285.79443786361537 hour

    """
    return (V_stock / Q_stock_max(Q_plant, C, tubing_color)).to(u.hr)


def M_stock(Q_plant, C, tubing_color, V_stock):
    """The mass of the material required to reach the desired stock
    concentration.

    Parameters
    ----------
    Q_plant : float
        flow rate of the plant

    C : float
        desired concentration within the plant

    tubing_color : string
        color of the tubing to be used

    V_stock : float
        volume of the stock container

    Returns
    -------
    float
        mass of material to be added to the stock container

    Examples
    --------
    >>> M_stock(7*u.mL/u.s, 100*u.NTU, "yellow-blue", 1*u.L)
    159.89684125188708 gram

    """
    return (C_stock_max(Q_plant, C, tubing_color) * V_stock).to(u.g)


def V_super_stock(Q_plant, C, tubing_color, V_stock, C_super_stock):
    """The volume of super stock added to the stock container to reach the
    desired concentration within the plant.

    Parameters
    ----------
    Q_plant : float
        flow rate of the plant

    C : float
        desired concentration within the plant

    tubing_color : string
        color of the tubing to be used

    V_stock : float
        volume of the stock container

    C_super_stock : float
        concentration of the super stock to be diluted down to the
        stock solution

    Returns
    -------
    float
        volume of super stock to be added to the stock container

    Examples
    --------
    >>> V_super_stock(7*u.mL/u.s, 2*u.mg/u.L, "orange-yellow", 1*u.L, 69.4*u.g/u.L)
    207.55100386637196 milliliter

    """
    C_stock = C_stock_max(Q_plant, C, tubing_color)
    return (V_stock * C_stock / C_super_stock).to(u.mL)


def Q_water(Q_plant, C_clay, C_pacl_min, tubing_clay, tubing_pacl):
    """This function calculates the required flow rate for water from the tap
    for the experiment.

    Parameters
    ----------
    Q_plant : float
        flow rate of the plant

    C_clay : float
        concentration of clay to be added, i.e. the desired influent turbidity

    C_pacl_min : float
        minimum coagulant dose of the mixed water in the flocculator

    tubing_clay
        color of the tubing to be used for the clay stock

    tubing_clay
        color of the tubing to be used for the PACL stock

    Returns
    -------
    float
        required flow rate for water for the experiment if it were coming from
        a tap

    Examples
    --------
    >>> Q_water(7*u.mL/u.s, 100*u.NTU, 0.2*u.mg/u.L, "yellow-blue", "orange-yellow")
    419.49514512465606 milliliter/minute

    """
    return (Q_plant - Q_stock_max(Q_plant, C_clay, tubing_clay) -
            Q_stock_max(Q_plant, C_pacl_min, tubing_pacl)).to(u.mL/u.min)


def pump_rpm(Q, tubing_color):
    """This function calculates the RPMs required for a given flow rate and
    tube color.

    Parameters
    ----------
    Q : float
        desired flow rate

    tubing_color
        color of the tubing to be used

    Returns
    -------
    float
        revolutions per minute to set the pump to for the desired flow rate

    Examples
    --------
    >>> pump_rpm(0.01*u.mL/u.s, "yellow-blue")
    4.031012804669423 rev/minute

    """
    flow_per_rev = Q6_roller(ID_colored_tube(tubing_color))
    return (Q / flow_per_rev).to(u.rev/u.min)



"""This file provides basic utility functions such as significant figures which
can be used throughout the plant design.

"""

# units allows us to include units in all of our calculations
import math

try:
    from aguaclara.core.units import unit_registry as u
except ModuleNotFoundError:
    from aguaclara.core.units import unit_registry as u

import numpy as np
import functools

#We need to fix the formatting so that it doesn't display trailing zeroes
#that are not significant.
def sig(x,n):
    """Return the 1st input reduced to a number of significant digits.

    x is a number that may include units. n is the number of significant
    digits to display.
    """
    # Check to see if the quantity x includes units so we can strip the
    # units and then reattach them at the end.
    if type(x) == type(1 * u.m):
        xunit = x.units
        xmag = float(x.magnitude)
        if n==1 and xmag>=1:
             req = round(xmag)
             return '{:~P}'.format(u.Quantity(req,xunit))

    else:
        xmag = x
    if xmag == 0.:
        return "0." + "0" * (n-1)
    if n == 1 and type(x)!=type(1 * u.m):
        return round(xmag)

    out = []
    if xmag < 0:
        out.append("-")
        xmag = -xmag
    e = int(math.log10(xmag))
    tens = math.pow(10, e - n + 1)
    y = math.floor(xmag / tens)
    if y < math.pow(10, n - 1):
        e = e -1
        tens = math.pow(10, e - n+1)
        y = math.floor(xmag / tens)
    if abs((y + 1.) * tens - xmag) <= abs(y * tens -xmag):
        y = y + 1
    if y >= math.pow(10,n):
        y = y / 10.
        e = e + 1

    m = "%.*g" % (n, y)
    if e < -2 or e >= n:
        out.append(m[0])
        if n > 1:
            out.append(".")
            out.extend(m[1 : n])
        out.append('e')
        if e > 0:
            out.append("+")
        out.append(str(e))
    elif e == (n - 1):
        out.append(m)
    elif e >= 0:
        out.append(m[:e + 1])
        if e+1 < len(m):
            out.append(".")
            out.extend(m[e + 1:])
    else:
        out.append("0.")
        out.extend(["0"] * -(e + 1))
        out.append(m)

    if type(x) == type(1 * u.m):
        req = "".join(out)
        return '{:~P}'.format(u.Quantity(req,xunit))
    else:
        return "".join(out)


def stepceil_with_units(param, step, unit):
    """This function returns the smallest multiple of 'step' greater than or
    equal to 'param' and outputs the result in Pint units.
    This function is unit-aware and functions without requiring translation
    so long as 'param' and 'unit' are of the same dimensionality.
    """
    counter = 0 * unit
    while counter < param.to(unit):
        counter += step * unit
    return counter


# Take the values of the array, compare to x, find the index of the first value less than or equal to x
def floor_nearest(x,array):
    myindex = np.argmax(array >= x) - 1
    return array[myindex]


# Take the values of the array, compare to x, find the index of the first value greater or equal to x
def ceil_nearest(x,array):
    myindex = np.argmax(array >= x)
    return array[myindex]


def list_handler(func):
    """Wraps a function to handle list inputs."""
    @functools.wraps(func)
    def wrapper(*args, HandlerResult="nparray", **kwargs):
        """Run through the wrapped function once for each array element.

        :param HandlerResult: output type. Defaults to numpy arrays.
        """
        sequences = []
        enumsUnitCheck = enumerate(args)
        argsList = list(args)
        #This for loop identifies pint unit objects and strips them
        #of their units.
        for num, arg in enumsUnitCheck:
            if type(arg) == type(1 * u.m):
                argsList[num] = arg.to_base_units().magnitude
        enumsUnitless = enumerate(argsList)
        #This for loop identifies arguments that are sequences and
        #adds their index location to the list 'sequences'.
        for num, arg in enumsUnitless:
            if isinstance(arg, (list, tuple, np.ndarray)):
                sequences.append(num)
        #If there are no sequences to iterate through, simply return
        #the function.
        if len(sequences) == 0:
            result = func(*args, **kwargs)
        else:
            #iterant keeps track of how many times we've iterated and
            #limiter stops the loop once we've iterated as many times
            #as there are list elements. Without this check, a few
            #erroneous runs will occur, appending the last couple values
            #to the end of the list multiple times.
            #
            #We only care about the length of sequences[0] because this
            #function is recursive, and sequences[0] is always the relevant
            #sequences for any given run.
            limiter = len(argsList[sequences[0]])
            iterant = 0
            result = []
            for num in sequences:
                for arg in argsList[num]:
                    if iterant >= limiter:
                        break
                    #We can safely replace the entire list argument
                    #with a single element from it because of the looping
                    #we're doing. We redefine the object, but that
                    #definition remains within this namespace and does
                    #not penetrate further up the function.
                    argsList[num] = arg
                    #Here we dive down the rabbit hole. This ends up
                    #creating a multi-dimensional array shaped by the
                    #sizes and shapes of the lists passed.
                    result.append(wrapper(*argsList,
                                          HandlerResult=HandlerResult, **kwargs))
                    iterant += 1
            #HandlerResult allows the user to specify what type to
            #return the generated sequence as. It defaults to numpy
            #arrays because functions tend to handle them better, but if
            #the user does not wish to import numpy the base Python options
            #are available to them.
            if HandlerResult == "nparray":
                result = np.array(result)
            elif HandlerResult == "tuple":
                result = tuple(result)
            elif HandlerResult == "list":
                result == list(result)
        return result
    return wrapper


def check_range(*args):
    """Check whether passed paramters fall within approved ranges.

    Does not return anything, but will raise an error if a parameter falls
    outside of its defined range.

    Input should be passed as an array of sequences, with each sequence
    having three elements:
        [0] is the value being checked,
        [1] is the range parameter(s) within which the value should fall, and
        [2] is the name of the parameter, for better error messages.
    If [2] is not supplied, "Input" will be appended as a generic name.

    Range requests that this function understands are listed in the
    knownChecks sequence.
    """
    knownChecks = ('>0', '>=0', '0-1', '<0', '<=0', 'int', 'boolean')
    for arg in args:
        #Converts arg to a mutable list
        arg = [*arg]
        if len(arg) == 1:
            #arg[1] details what range the parameter should fall within; if
            #len(arg) is 1 that means a validity was not specified and the
            #parameter should not have been passed in its current form
            raise TypeError("No range-validity parameter provided.")
        elif len(arg) == 2:
            #Appending 'Input" to the end allows us to give more descriptive
            #error messages that do not fail if no description was supplied.
            arg.append("Input")
        #This ensures that all whitespace is removed before checking if the
        #request is understood
        arg[1] = "".join(arg[1].lower().split())
        #This block checks that each range request is understood.
        #If the request is a compound one, it must be separated into individual
        #requests for validity comprehension
        for i in arg[1].split(","):
            if i not in knownChecks:
                raise RuntimeError("Unknown parameter validation "
                                       "request: {0}.".format(i))
        if not isinstance(arg[0], (list, tuple, np.ndarray)):
            arg[0] = [arg[0]]
        for i in arg[0]:
            if '>0' in arg[1] and i <= 0:
                raise ValueError("{1} is {0} but must be greater than "
                                 "0.".format(i, arg[2]))
            if '>=0' in arg[1] and i <0:
                raise ValueError("{1} is {0} but must be 0 or "
                                 "greater.".format(i, arg[2]))
            if '0-1' in arg[1] and not 0 <= i <= 1:
                raise ValueError("{1} is {0} but must be between 0 and "
                                 "1.".format(i, arg[2]))
            if '<0' in arg[1] and i >= 0:
                raise ValueError("{1} is {0} but must be less than "
                                 "0.".format(i, arg[2]))
            if '<=0' in arg[1] and i >0:
                raise ValueError("{1} is {0} but must be 0 or "
                                 "less.".format(i, arg[2]))
            if 'int' in arg[1] and int(i) != i:
                raise TypeError("{1} is {0} but must be a numeric "
                                "integer.".format(i, arg[2]))
            if 'boolean' in arg[1] and type(i) != bool:
                raise TypeError("{1} is {0} but must be a "
                                "boolean.".format(i, arg[2]))



"""This file contains the default values which may be overriden by user inputs
for design parameters of AguaClara plants.

"""
from aguaclara.core.units import unit_registry as u

## ETLF

# Entrance Tank
L_ET_MAX = 2.2 * u.m

# LFOM
HL_LFOM = 20 * u.cm

S_LFOM_ORIFICE = 1 * u.cm  # minimum wall distance between orifices, for lfom structural stability

# Flocculator
HL_FLOC = 0.4 * u.m

COLL_POT = 37000  # collision potential, also referred to as Gt

FREEBOARD = 10 * u.cm

# Sedimentation tank
THICKNESS_PLANT_FLOOR = 0.2 * u.m  # plant floor slab thickness

THICKNESS_SED_WALL = 0.15 * u.m  # thickness of sed tank dividing wall

FLOC_BLANKET_HEIGHT = 0.25 * u.m  # vertical height of floc blanket from peak of slope to weir

HL_OUTLET_MAN = 4 * u.cm  # head loss through the outlet manifold



"""
General purpose utility library..
"""

import math


class Constants(object):
    """
    Constants.
    """

    class Earth(object):
        """
        Earth constants.
        """
        GRAVITY = 9.80665

    class Mars(object):
        """
        Mars constants.
        """
        GRAVITY = 9


def orbital_speed(planet, planatary_radius, altitude):
    """
    Calculate the orbital speed of an object.

    v = R*sqrt(g/(R+h))

    >>> from math import *
    >>> round(orbital_speed(Constants.Earth, 600000, 70), 3)
    2425.552
    """
    total_altitude = planatary_radius + altitude
    speed = planatary_radius * \
        math.sqrt(planet.GRAVITY / total_altitude)
    return speed


def circumference(radius):
    """
    Calculate the circumference of an object.

    2*pi*r

    >>> from math import *
    >>> round(circumference(600000), 3)
    3769911.184
    """
    distance = 2 * math.pi * radius

    return distance


def orbital_period(planet, planatary_radius, altitude):
    """
    Calculate the orbital period of an object.

    d = v*t

    >>> from math import *
    >>> round(orbital_period(Constants.Earth, 600000, 70), 3)
    1554.43
    """
    cir = circumference(planatary_radius + altitude)

    return cir / orbital_speed(planet, planatary_radius, altitude)



"""This utility is used to provide useful functions for all things related to
pipeline design.

"""

from aguaclara.core.units import unit_registry as u
from aguaclara.core import physchem as pc
import aguaclara.core.constants as con
import aguaclara.core.materials_database as mats
import numpy as numpy

@u.wraps(u.m**3/u.s, [u.m, u.m, None, u.m], False)
def flow_pipeline(diameters: numpy.ndarray, lengths: numpy.ndarray, k_minors: numpy.ndarray, target_headloss: float,
                  nu=con.NU_WATER, pipe_rough=mats.PIPE_ROUGH_PVC):
    """
    This function takes a single pipeline with multiple sections, each potentially with different diameters,
    lengths and minor loss coefficients and determines the flow rate for a given headloss.

    Args:
        diameters: list of diameters, where the i_th diameter corresponds to the i_th pipe section
        lengths: list of diameters, where the i_th diameter corresponds to the i_th pipe section
        k_minors: list of diameters, where the i_th diameter corresponds to the i_th pipe section
        target_headloss: a single headloss describing the total headloss through the system
        nu: The fluid dynamic viscosity of the fluid. Defaults to water at room temperature (1 * 10**-6 * m**2/s)
        pipe_rough:  The pipe roughness. Defaults to PVC roughness.
    Returns:
        flow: the total flow through the system
    """

    # Ensure all the arguments except total headloss are the same length
    #TODO

    # Total number of pipe lengths
    n = diameters.size

    # Start with a flow rate guess based on the flow through a single pipe section
    flow = pc.flow_pipe(diameters[0], target_headloss, lengths[0], nu, pipe_rough, k_minors[0])
    err = 1.0

    # Add all the pipe length headlosses together to test the error
    while abs(err) > 0.01 :
        headloss = sum([pc.headloss(flow, diameters[i], lengths[i], nu, pipe_rough,
                                    k_minors[i]).to(u.m).magnitude for i in range(n)])
        # Test the error. This is always less than one.
        err = (target_headloss - headloss) / (target_headloss + headloss)
        # Adjust the total flow in the direction of the error. If there is more headloss than target headloss,
        # The flow should be reduced, and vice-versa.
        flow = flow + err * flow

    return flow



from aguaclara.core.units import unit_registry as u
from scipy import special
from scipy.optimize import curve_fit
import collections
import pandas as pd
import numpy as np

# Carbonates
# The following code defines the carbonate system and provides functions for
# calculating Acid Neutralizing Capacity.
Kw = 10**(-14) * (u.mole/u.L)**2
K1_carbonate = 10**(-6.37)*u.mol/u.L
K2_carbonate = 10**(-10.25)*u.mol/u.L
K_Henry_CO2 = 10**(-1.5) * u.mole/(u.L*u.atm)
P_CO2 = 10**(-3.5) * u.atm


def invpH(pH):
    """This function calculates inverse pH

    Parameters
    ----------
    pH : float
        pH to be inverted

    Returns
    -------
    The inverse pH (in moles per liter) of the given pH

    Examples
    --------
    >>> invpH(8.25)
    5.623413251903491e-09 mole/liter
    >>> invpH(10)
    1e-10 mole/liter

    """
    return 10**(-pH)*u.mol/u.L


def alpha0_carbonate(pH):
    """This function calculates the fraction of total carbonates of the form
    H2CO3

    Parameters
    ----------
    pH : float
        pH of the system

    Returns
    -------
    fraction of CT in the form H2CO3

    Examples
    --------
    >>> alpha0_carbonate(8.25)
    0.01288388583402879 dimensionless
    >>> alpha0_carbonate(10)
    0.00015002337123256595 dimensionless

    """
    alpha0_carbonate = 1/(1+(K1_carbonate/invpH(pH)) *
                            (1+(K2_carbonate/invpH(pH))))
    return alpha0_carbonate


def alpha1_carbonate(pH):
    """This function calculates the fraction of total carbonates of the form
    HCO3-

    Parameters
    ----------
    pH : float
        pH of the system

    Returns
    -------
    fraction of CT in the form HCO3-

    Examples
    --------
    >>> alpha1_carbonate(8.25)
    0.9773426872930407 dimensionless
    >>> alpha1_carbonate(10)
    0.6399689750938067 dimensionless

    """
    alpha1_carbonate = 1/((invpH(pH)/K1_carbonate) + 1 +
                          (K2_carbonate/invpH(pH)))
    return alpha1_carbonate


def alpha2_carbonate(pH):
    """This function calculates the fraction of total carbonates of the form
    CO3-2

    Parameters
    ----------
    pH : float
        pH of the system

    Returns
    -------
    fraction of CT in the form CO3-2

    Examples
    --------
    >>> alpha2_carbonate(8.25)
    0.009773426872930407 dimensionless
    >>> alpha2_carbonate(10)
    0.35988100153496067 dimensionless

    """
    alpha2_carbonate = 1/(1+(invpH(pH)/K2_carbonate) *
                            (1+(invpH(pH)/K1_carbonate)))
    return alpha2_carbonate


def ANC_closed(pH, Total_Carbonates):
    """Acid neutralizing capacity (ANC) calculated under a closed system where
    there are no carbonates exchanged with the atmosphere during the
    experiment. Based on pH and total carbonates in the system.

    Parameters
    ----------
    pH : float
        pH of the system

    Total_Carbonates
        total carbonates in the system (mole/L)

    Returns
    -------
    The acid neutralizing capacity of the closed system (eq/L)

    Examples
    --------
    >>> ANC_closed(8.25, 1*u.mol/u.L)
    0.9968913136948984 equivalents/liter
    >>> ANC_closed(10, 1*u.mol/u.L)
    1.359830978063728 equivalents/liter

    """
    return (Total_Carbonates * (u.eq/u.mol * alpha1_carbonate(pH) +
            2 * u.eq/u.mol * alpha2_carbonate(pH)) +
            1 * u.eq/u.mol * Kw/invpH(pH) - 1 * u.eq/u.mol * invpH(pH))


def ANC_open(pH):
    """Acid neutralizing capacity (ANC) calculated under an open system, based
    on pH.

    Parameters
    ----------
    pH : float
        pH of the system

    Returns
    -------
    The acid neutralizing capacity of the closed system (eq/L)

    Examples
    --------
    >>> ANC_open(8.25)
    0.0007755217825265541 equivalents/liter
    >>> ANC_open(10)
    0.09073461016054905 equivalents/liter

    """
    return ANC_closed(pH, P_CO2*K_Henry_CO2/alpha0_carbonate(pH))


def aeration_data(DO_column, dirpath):
    """This function extracts the data from folder containing tab delimited
    files of aeration data. The file must be the original tab delimited file.
    All text strings below the header must be removed from these files.
    The file names must be the air flow rates with units of micromoles/s.
    An example file name would be "300.xls" where 300 is the flowr ate in
    micromoles/s. The function opens a file dialog for the user to select
    the directory containing the data.

    Parameters
    ----------
    DO_column : int
        index of the column that contains the dissolved oxygen concentration
        data.

    dirpath : string
        path to the directory containing aeration data you want to analyze

    Returns
    -------
    filepaths : string list
        all file paths in the directory sorted by flow rate

    airflows : numpy array
        sorted array of air flow rates with units of micromole/s attached

    DO_data : numpy array list
        sorted list of numpy arrays. Thus each of the numpy data arrays can
        have different lengths to accommodate short and long experiments

    time_data : numpy array list
        sorted list of numpy arrays containing the times with units of seconds

    Examples
    --------

    """
    #return the list of files in the directory
    filenames = os.listdir(dirpath)
    #extract the flowrates from the filenames and apply units
    airflows = ((np.array([i.split('.', 1)[0] for i in filenames])).astype(np.float32))
    #sort airflows and filenames so that they are in ascending order of flow rates
    idx = np.argsort(airflows)
    airflows = (np.array(airflows)[idx])*u.umole/u.s
    filenames = np.array(filenames)[idx]

    filepaths = [os.path.join(dirpath, i) for i in filenames]
    #DO_data is a list of numpy arrays. Thus each of the numpy data arrays can have different lengths to accommodate short and long experiments
    # cycle through all of the files and extract the column of data with oxygen concentrations and the times
    DO_data=[Column_of_data(i,0,-1,DO_column,'mg/L') for i in filepaths]
    time_data=[(ftime(i,0,-1)).to(u.s) for i in filepaths]
    aeration_collection = collections.namedtuple('aeration_results','filepaths airflows DO_data time_data')
    aeration_results = aeration_collection(filepaths, airflows, DO_data, time_data)
    return aeration_results


def O2_sat(P_air, temp):
    """This equation returns saturaed oxygen concentration in mg/L. It is valid
    for 278 K < T < 318 K

    Parameters
    ----------
    Pressure_air : float
        air pressure with appropriate units.
    Temperature :
        water temperature with appropriate units

    Returns
    -------
    Saturated oxygen concentration in mg/L

    Examples
    --------
    >>> O2_sat(1*u.atm , 300*u.kelvin)
    8.093157231428425 milligram/liter

    """
    fraction_O2 = 0.21
    P_O2 = P_air * fraction_O2
    return ((P_O2.to(u.atm).magnitude) *
            u.mg/u.L*np.exp(1727 / temp.to(u.K).magnitude - 2.105))


def Gran(data_file_path):
    """This function extracts the data from a ProCoDA Gran plot file.
    The file must be the original tab delimited file.

    Parameters
    ----------
    data_file_path : string
        File path. If the file is in the working directory, then the file name
        is sufficient.

    Returns
    -------
    V_titrant : float
        volume of titrant in mL

    ph_data : numpy array
        pH of the sample

    V_sample : float
        volume of the original sample that was titrated in mL

    Normality_titrant : float
        normality of the acid used to titrate the sample in mole/L

    V_equivalent : float
        volume of acid required to consume all of the ANC in mL

    ANC : float
        Acid Neutralizing Capacity of the sample in mole/L

    Examples
    --------

    """
    df = pd.read_csv(data_file_path, delimiter='\t', header=5)
    V_t = np.array(pd.to_numeric(df.iloc[0:, 0]))*u.mL
    pH = np.array(pd.to_numeric(df.iloc[0:, 1]))
    df = pd.read_csv(data_file_path, delimiter='\t', header=-1, nrows=5)
    V_S = pd.to_numeric(df.iloc[0, 1])*u.mL
    N_t = pd.to_numeric(df.iloc[1, 1])*u.mole/u.L
    V_eq = pd.to_numeric(df.iloc[2, 1])*u.mL
    ANC_sample = pd.to_numeric(df.iloc[3, 1])*u.mole/u.L
    Gran_collection = collections.namedtuple('Gran_results', 'V_titrant ph_data V_sample Normality_titrant V_equivalent ANC')
    Gran = Gran_collection(V_titrant=V_t, ph_data=pH, V_sample=V_S,
                           Normality_titrant=N_t, V_equivalent=V_eq,
                           ANC=ANC_sample)
    return Gran


# Reactors
# The following code is for reactor responses to tracer inputs.
def CMFR(t, C_initial, C_influent):
    """This function calculates the output concentration of a completely mixed
    flow reactor given an influent and initial concentration.

    Parameters
    ----------
    C_initial : float
        The concentration in the CMFR at time zero.

    C_influent : float
        The concentration entering the CMFR.

    t : float (array)
        time made dimensionless by dividing by the residence time of the CMFR.
        It can be a single value or a numpy array.

    Returns
    -------
    float
        Effluent concentration

    Examples
    --------
    >>> CMFR(0.1, 0*u.mg/u.L, 10*u.mg/u.L)
    0.9516258196404048 milligram/liter
    >>> CMFR(0.9, 5*u.mg/u.L, 10*u.mg/u.L)
    7.967151701297004 milligram/liter

    """
    return C_influent * (1-np.exp(-t)) + C_initial*np.exp(-t)


def E_CMFR_N(t, N):
    """This function calculates a dimensionless measure of the output tracer
    concentration from a spike input to a series of completely mixed flow
    reactors.

    Parameters
    ----------
    t : float (array)
        time made dimensionless by dividing by the reactor residence time.
        t can be a single value or a numpy array.

    N : float
        number of completely mixed flow reactors (CMFRS) in series.
        This would logically be constrained to real numbers greater than 1.

    Returns
    -------
    float
        dimensionless measure of the output tracer concentration
        (Concentration * volume of 1 CMFR) / (mass of tracer)

    Examples
    --------
    >>> E_CMFR_N(0.5, 3)
    0.7530642905009506
    >>> E_CMFR_N(0.1, 1)
    0.9048374180359595

    """
    return (N**N)/special.gamma(N) * (t**(N-1))*np.exp(-N*t)


def E_Advective_Dispersion(t, Pe):
    """This function calculates a dimensionless measure of the output tracer
    concentration from a spike input to reactor with advection and dispersion.
    Parameters
    ----------
    t : float (array)
        time made dimensionless by dividing by the reactor residence time.
        t can be a single value or a numpy array.

    Pe : float
        The ratio of advection to dispersion
        ((mean fluid velocity)/(Dispersion*flow path length))

    Returns
    -------
    float
        dimensionless measure of the output tracer concentration
        (Concentration * volume of reactor) / (mass of tracer)

    Examples
    --------
    >>> E_Advective_Dispersion(0.5, 5)
    0.47748641153355664

    """
    # replace any times at zero with a number VERY close to zero to avoid
    # divide by zero errors
    if isinstance(t, list):
        t[t == 0] = 10**(-50)
    return (Pe/(4*np.pi*t))**(0.5)*np.exp((-Pe*((1-t)**2))/(4*t))


def Tracer_CMFR_N(t_seconds, t_bar, C_bar, N):
    """Used by Solver_CMFR_N. All inputs and outputs are unitless. This is
    The model function, f(x, ...). It takes the independent variable as the
    first argument and the parameters to fit as separate remaining arguments.

    Parameters
    ----------
    t_seconds : float list
        Array of times

    t_bar : float
        Average time spent in the reactor

    C_bar : float
        Average concentration.
        (Mass of tracer)/(volume of the reactor)

    N : float
        number of completely mixed flow reactors (CMFRS) in series.
        This would logically be constrained to real numbers greater than 1.

    Returns
    -------
    float list
        The model concentration as a function of time

    Examples
    --------
    >>> Tracer_CMFR_N([1, 2, 3, 4, 5]*u.s, 5*u.s, 10*u.mg/u.L, 3)
    \\[\\begin{pmatrix}2.963582834907743 & 6.505794977303565 & 8.033525967569107 & 7.838031164205239 & 6.721254229661633\\end{pmatrix} milligram/liter\\]

    """
    return C_bar*E_CMFR_N(t_seconds/t_bar, N)


def Solver_CMFR_N(t_data, C_data, theta_guess, C_bar_guess):
    """Use non-linear least squares to fit the function
    Tracer_CMFR_N(t_seconds, t_bar, C_bar, N), to reactor data.

    Parameters
    ----------
    t_data : float list
        Array of times with units

    C_data : float list
        Array of tracer concentration data with units

    theta_guess : float
        Estimate of time spent in one CMFR with units.

    C_bar_guess : float
        Estimate of average concentration with units
        (Mass of tracer)/(volume of one CMFR)

    Returns
    -------
    tuple
        theta : float
            residence time in seconds

        C_bar : float
            average concentration with same units as C_bar_guess

        N : float
            number of CMFRS in series that best fit the data

    """
    C_unitless = C_data.magnitude
    C_units = str(C_bar_guess.units)
    t_seconds = (t_data.to(u.s)).magnitude
    # assume that a guess of 1 reactor in series is close enough to get a solution
    p0 = [theta_guess.to(u.s).magnitude, C_bar_guess.magnitude,1]
    popt, pcov = curve_fit(Tracer_CMFR_N, t_seconds, C_unitless, p0)
    Solver_theta = popt[0]*u.s
    Solver_C_bar = popt[1]*u(C_units)
    Solver_N = popt[2]
    Reactor_results = collections.namedtuple('Reactor_results','theta C_bar N')
    CMFR = Reactor_results(theta=Solver_theta, C_bar=Solver_C_bar, N=Solver_N)
    return CMFR


def Tracer_AD_Pe(t_seconds, t_bar, C_bar, Pe):
    """Used by Solver_AD_Pe. All inputs and outputs are unitless. This is the
    model function, f(x, ...). It takes the independent variable as the
    first argument and the parameters to fit as separate remaining arguments.

    Parameters
    ----------
    t_seconds : float list
        Array of times

    t_bar : float
        Average time spent in the reactor

    C_bar : float
        Average concentration.
        (Mass of tracer)/(volume of the reactor)

    Pe : float
        The Peclet number for the reactor.

    Returns
    -------
    float
        The model concentration as a function of time

    Examples
    --------
    >>> Tracer_AD_Pe([1, 2, 3, 4, 5]*u.s, 5*u.s, 10*u.mg/u.L, 5)
    \\[\\begin{pmatrix}0.2583373169261504 & 3.237939891647294 & 5.834983303390744 & 6.625088308600714 & 6.307831305050401\\end{pmatrix} milligram/liter\\]

    """
    return C_bar*E_Advective_Dispersion(t_seconds/t_bar, Pe)


def Solver_AD_Pe(t_data, C_data, theta_guess, C_bar_guess):
    """Use non-linear least squares to fit the function
    Tracer_AD_Pe(t_seconds, t_bar, C_bar, Pe) to reactor data.

    Parameters
    ----------
    t_data : float list
        Array of times with units

    C_data : float list
        Array of tracer concentration data with units

    theta_guess : float
        Estimate of time spent in one CMFR with units.

    C_bar_guess : float
        Estimate of average concentration with units
        (Mass of tracer)/(volume of one CMFR)

    Returns
    -------
    tuple
        theta : float
            residence time in seconds

        C_bar : float
            average concentration with same units as C_bar_guess

        Pe : float
            peclet number that best fits the data

    """

    C_unitless = C_data.magnitude
    C_units = str(C_bar_guess.units)
    t_seconds = (t_data.to(u.s)).magnitude
    # assume that a guess of 1 reactor in series is close enough to get a solution
    p0 = [theta_guess.to(u.s).magnitude, C_bar_guess.magnitude,5]
    popt, pcov = curve_fit(Tracer_AD_Pe, t_seconds, C_unitless, p0)
    Solver_theta = popt[0]*u.s
    Solver_C_bar = popt[1]*u(C_units)
    Solver_Pe = popt[2]
    Reactor_results = collections.namedtuple('Reactor_results', 'theta C_bar Pe')
    AD = Reactor_results(theta=Solver_theta, C_bar=Solver_C_bar, Pe=Solver_Pe)
    return AD



"""This file is used to calculate the design paramters for the chemical dose
controller of an AguaClara plant.

"""

import numpy as np

from aguaclara.core import physchem as pc, utility as ut

from aguaclara.core.units import unit_registry as u


#==============================================================================
# Functions for Coagulant Viscosities and Selecting Available Tube Diameters
#==============================================================================

def _DiamTubeAvail(en_tube_series = True):
    if en_tube_series:
        return 1*u.mm
    else:
        return (1/16)*u.inch
# This section may be unnecessary
#==============================================================================
# def _DiamTubeAvail(en_tube_series = True):
#     if en_tube_series:
#         return 1*u.mm
#     else:
#         return (1/16)*u.inch
#==============================================================================

@u.wraps(u.m**2/u.s, [u.kg/u.m**3, u.degK], False)
def viscosity_kinematic_alum(conc_alum, temp):
    """Return the dynamic viscosity of water at a given temperature.

    If given units, the function will automatically convert to Kelvin.
    If not given units, the function will assume Kelvin.
    This function assumes that the temperature dependence can be explained
    based on the effect on water and that there is no confounding effect from
    the coagulant.
    """
    nu = (1 + (4.255 * 10**-6) * conc_alum**2.289) * pc.viscosity_kinematic(temp).magnitude
    return nu


@u.wraps(u.m**2/u.s, [u.kg/u.m**3, u.degK], False)
def viscosity_kinematic_pacl(conc_pacl, temp):
    """Return the dynamic viscosity of water at a given temperature.

    If given units, the function will automatically convert to Kelvin.
    If not given units, the function will assume Kelvin.
    This function assumes that the temperature dependence can be explained
    based on the effect on water and that there is no confounding effect from
    the coagulant.
    """
    nu = (1 + (2.383 * 10**-5) * conc_pacl**1.893) * pc.viscosity_kinematic(temp).magnitude
    return nu


@u.wraps(u.m**2/u.s, [u.kg/u.m**3, u.degK, None], False)
def viscosity_kinematic_chem(conc_chem, temp, en_chem):
     """Return the dynamic viscosity of water at a given temperature.

    If given units, the function will automatically convert to Kelvin.
    If not given units, the function will assume Kelvin.
    """
     if en_chem == 0:
         nu = viscosity_kinematic_alum(conc_chem, temp).magnitude
     if en_chem == 1:
         nu =  viscosity_kinematic_pacl(conc_chem, temp).magnitude
     if en_chem not in [0,1]:
         nu =  pc.viscosity_kinematic(temp).magnitude
     return nu



#==============================================================================
# Flow rate Constraints for Laminar Tube Flow, Deviation from Linear Head Loss
# Behavior, and Lowest Possible Flow
#==============================================================================


@u.wraps(u.m**3/u.s, [u.m, u.m, None, None], False)
def max_linear_flow(Diam, HeadlossCDC, Ratio_Error, KMinor):
    """Return the maximum flow that will meet the linear requirement.
    Maximum flow that can be put through a tube of a given diameter without
    exceeding the allowable deviation from linear head loss behavior
    """
    flow = (pc.area_circle(Diam)).magnitude * np.sqrt((2 * Ratio_Error * HeadlossCDC * pc.gravity)/ KMinor)
    return flow.magnitude




#==============================================================================
# Length of Tubing Required Given Head Loss, Max Flow, and Diameter
#==============================================================================

# Length of tube required to get desired head loss at maximum flow based on
# the Hagen-Poiseuille equation.
@u.wraps(u.m, [u.m**3/u.s, u.m, u.m, u.kg/u.m**3, u.degK, None, None], False)
def _len_tube(Flow, Diam, HeadLoss, conc_chem, temp, en_chem, KMinor):
    """Length of tube required to get desired head loss at maximum flow based on
    the Hagen-Poiseuille equation."""
    num1 = pc.gravity.magnitude * HeadLoss * np.pi * (Diam**4)
    denom1 = 128 * viscosity_kinematic_chem(conc_chem, temp, en_chem) * Flow
    num2 = Flow * KMinor
    denom2 = 16 * np.pi * viscosity_kinematic_chem(conc_chem, temp, en_chem)
    len = ((num1/denom1) - (num2/denom2))
    return len.magnitude



#==============================================================================
# Helper Functions
#==============================================================================
@u.wraps(None, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, None, None], False)
def _n_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                  DiamTubeAvail, HeadlossCDC, Ratio_Error, KMinor):

    np.ceil((FlowPlant * ConcDoseMax
               )/ ConcStock*max_linear_flow(DiamTubeAvail, HeadlossCDC, Ratio_Error, KMinor).magnitude)
    return np.ceil((FlowPlant * ConcDoseMax) /
            (ConcStock * max_linear_flow(DiamTubeAvail, HeadlossCDC, Ratio_Error, KMinor).magnitude))


@u.wraps(u.m**3/u.s, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3], False)
def _flow_chem_stock(FlowPlant, ConcDoseMax, ConcStock):
    FlowPlant * ConcDoseMax / ConcStock
    return FlowPlant * ConcDoseMax / ConcStock


@u.wraps(u.m**3/u.s, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m,None,None], False)
def _flow_cdc_tube(FlowPlant, ConcDoseMax, ConcStock,
                   DiamTubeAvail, HeadlossCDC,Ratio_Error, KMinor):

    (_flow_chem_stock(FlowPlant, ConcDoseMax, ConcStock)
     ) / (_n_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                  DiamTubeAvail, HeadlossCDC,Ratio_Error, KMinor))
    return (_flow_chem_stock(FlowPlant, ConcDoseMax, ConcStock).magnitude
            ) / (_n_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                        DiamTubeAvail, HeadlossCDC,Ratio_Error, KMinor))



#
@u.wraps(u.m, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, u.degK, None, None], False)
def _length_cdc_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                           DiamTubeAvail, HeadlossCDC, temp, en_chem, KMinor):
    """Calculate the length of each diameter tube given the corresponding flow rate
    and coagulant. Choose the tube that is shorter than the maximum length tube."""

    Flow = _flow_cdc_tube(FlowPlant, ConcDoseMax, ConcStock, DiamTubeAvail, HeadlossCDC,Ratio_Error, KMinor).magnitude


    return _len_tube(Flow, DiamTubeAvail, HeadlossCDC, ConcStock, temp, en_chem, KMinor).magnitude


# Find the index of that tube
@u.wraps(None, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, u.m, u.degK, None, None], False)
def i_cdc(FlowPlant, ConcDoseMax, ConcStock,
          DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp,
          en_chem, KMinor):

    tube_array = (_length_cdc_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                                        DiamTubeAvail, HeadlossCDC, temp, en_chem,
                                        KMinor)).magnitude


    if tube_array[0] < float(LenCDCTubeMax):
        y = ut.floor_nearest(LenCDCTubeMax,tube_array)
        myindex =np.where(tube_array==y)[0][0]

    else:
        myindex = 0

    return myindex



#==============================================================================
# Final easy to use functions
#==============================================================================

@u.wraps(u.m, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, u.m, u.degK, None, None], False)
def len_cdc_tube(FlowPlant, ConcDoseMax, ConcStock,
                 DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp,
                 en_chem, KMinor):
   """The length of tubing may be longer than the max specified if the stock
   concentration is too high to give a viable solution with the specified
   length of tubing."""
   index = i_cdc(FlowPlant, ConcDoseMax, ConcStock,
                DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp,
                en_chem, KMinor)
   len_cdc_tube = (_length_cdc_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                                        DiamTubeAvail, HeadlossCDC, temp, en_chem,
                                        KMinor))[index].magnitude

   return len_cdc_tube


@u.wraps(u.m, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, u.m, u.degK, None, None], False)
def diam_cdc_tube(FlowPlant, ConcDoseMax, ConcStock,
                  DiamTubeAvail, HeadlossCDC, LenCDCTubeMax,
                  temp, en_chem, KMinor):

    index = i_cdc(FlowPlant, ConcDoseMax, ConcStock,
                   DiamTubeAvail, HeadlossCDC, LenCDCTubeMax,
                   temp, en_chem, KMinor)

    diam_cdc_tube = DiamTubeAvail[index]

    return diam_cdc_tube


@u.wraps(None, [u.m**3/u.s, u.kg/u.m**3, u.kg/u.m**3, u.m, u.m, u.m, u.degK, None, None], False)
def n_cdc_tube(FlowPlant, ConcDoseMax, ConcStock,
          DiamTubeAvail, HeadlossCDC, LenCDCTubeMax,
          temp, en_chem, KMinor):

    index = i_cdc(FlowPlant, ConcDoseMax, ConcStock,
                  DiamTubeAvail, HeadlossCDC, LenCDCTubeMax,
                  temp, en_chem, KMinor)

    n_cdc_tube = _n_tube_array(FlowPlant, ConcDoseMax, ConcStock,
                  DiamTubeAvail, HeadlossCDC, Ratio_Error, KMinor)[index]

    return n_cdc_tube




# testing
FlowPlant = 100*u.L/u.s
DiamTubeAvail = np.array(np.arange(1/16,6/16,1/16))*u.inch
temp = u.Quantity(20,u.degC)
HeadlossCDC = 20*(u.cm)
ConcStock = 51.4*(u.gram/u.L)
ConcDoseMax = 2*(u.mg/u.L)
LenCDCTubeMax = 4 * u.m
en_chem = 2
KMinor = 2
Ratio_Error=0.1
x=len_cdc_tube(FlowPlant, ConcDoseMax, ConcStock,
                 DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp,
                 en_chem, KMinor)
#print(x)
#print(diam_cdc_tube(FlowPlant, ConcDoseMax, ConcStock, DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp, en_chem, KMinor).to(u.inch))
#print(n_cdc_tube(FlowPlant, ConcDoseMax, ConcStock, DiamTubeAvail, HeadlossCDC, LenCDCTubeMax, temp, en_chem, KMinor))



import unittest
import numpy as np

from aguaclara.core import materials_database as mat

from aguaclara.core.units import unit_registry as u

from aguaclara.unit_process_design import lfom


class LfomTest(unittest.TestCase):
    def test_width_stout(self):
        """"width_stout should give known result with known input.
        Test cases were calculated using outputs from original Mathcad code.

        """
        checks = ((20 * u.cm, 1 * u.cm, 11.408649616179787 * u.s/u.m**2),
                  (40 * u.cm, 40 * u.cm, 0.9019329453483474 * u.s/u.m**2))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(lfom.width_stout(i[0], i[1]), i[2], places=3)

    def test_n_lfom_rows(self):
        """"n_rows should give known result with known input.
        Test cases were calculated using outputs from original Mathcad code.

        """
        checks = ((60 * u.L/u.s, 20 * u.cm, 4),
                  (60 * u.L/u.s, 40 * u.cm, 8),
                  (20 * u.L/u.s, 20 * u.cm, 8),
                  (1 * u.L/u.s, 20 * u.cm, 8),
                  (1 * u.L/u.s, 40 * u.cm, 8))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(lfom.n_lfom_rows(i[0], i[1]), i[2])

    def test_dist_center_lfom_rows(self):
        """b_rows should give known result with known input.
        Test cases were calculated using outputs from original Mathcad code.

        """
        checks = ((60 * u.L/u.s, 20 * u.cm, 0.05 * u.m),
                  (20 * u.L/u.s, 20 * u.cm, 0.025 * u.m),
                  (1 * u.L/u.s, 40 * u.cm, 0.05 * u.m))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(lfom.dist_center_lfom_rows(i[0], i[1]), i[2])

    def test_vel_lfom_pipe_critical(self):
        """vel_critical should give known result with known input.
        Test cases were calculated using outputs from original Mathcad code.

        """
        checks = ((20 * u.cm, 0.8405802802312778 * u.m/u.s),
                  (40 * u.cm, 1.18876003256645 * u.m/u.s))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(lfom.vel_lfom_pipe_critical(i[0]), i[1])

    def test_area_lfom_pipe_min(self):
        """area_lfom_pipe_min should give known result with known input.
        Test cases were calculated using outputs from original Mathcad code.

        """
        checks = ((60 * u.L/u.s, 20 * u.cm, 0.10706889290245702 * u.m**2),
                  (60 * u.L/u.s, 40 * u.cm, 0.07570914022546356 * u.m**2),
                  (20 * u.L/u.s, 20 * u.cm, 0.035689630967485675 * u.m**2),
                  (20 * u.L/u.s, 40 * u.cm, 0.02523638007515452 * u.m**2))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(lfom.area_lfom_pipe_min(i[0], i[1]), i[2])

    def test_nom_diam_lfom_pipe(self):
        """nom_diam_lfom_pipe should give known result with known input.
        Test cases were calculated using outputs from original Mathcad code and
        the nominal diameter function written in pipedatabase which has already
        been tested.

        """
        checks = ((60 * u.L/u.s, 20 * u.cm, 16.0 * u.inch),
                  (60 * u.L/u.s, 40 * u.cm, 16.0 * u.inch),
                  (20 * u.L/u.s, 20 * u.cm, 10.0 * u.inch),
                  (20 * u.L/u.s, 40 * u.cm, 8.0 * u.inch))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(lfom.nom_diam_lfom_pipe(i[0], i[1]), i[2])

##### Below here is not finished
    def test_area_lfom_orifices_top(self):
        """area_lfom_orifices_top should give known result with known input.
        Test cases were calculated using outputs from original Mathcad code and
        the nominal diameter function written in pipedatabase which has already
        been tested.

        """
        checks = ((60 * u.L/u.s, 20 * u.cm, 0.008181566649077958 * u.m**2),
                  (60 * u.L/u.s, 40 * u.cm, 0.0027945370213839633 * u.m**2),
                  (20 * u.L/u.s, 20 * u.cm, 0.0013173573853983045 * u.m**2),
                  (20 * u.L/u.s, 40 * u.cm, 0.0009315123404613211 * u.m**2))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(lfom.area_lfom_orifices_top(i[0], i[1]), i[2])

    def test_d_lfom_orifices_max(self):
        """d_lfom_orifices_max should give known result with known input.
        Test cases were calculated using outputs from original Mathcad code and
        the nominal diameter function written in pipedatabase which has already
        been tested.

        """
        checks = ((60 * u.L/u.s, 20 * u.cm,  0.10206416704942245* u.m),
                  (60 * u.L/u.s, 40 * u.cm,  0.05964993750920847* u.m),
                  (20 * u.L/u.s, 20 * u.cm,  0.04095499380586013* u.m),
                  (20 * u.L/u.s, 40 * u.cm,  0.03443890747808586* u.m))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(lfom.d_lfom_orifices_max(i[0], i[1]), i[2])

    def test_orifice_diameter(self):
        """orifice_diameter should give known result with known input.
        Test cases were calculated using outputs from original Mathcad code and
        the nominal diameter function written in pipedatabase which has already
        been tested.

        """
        checks = ((60 * u.L/u.s, 20 * u.cm, mat.DIAM_DRILL_ENG, 1* u.m),
                  (60 * u.L/u.s, 40 * u.cm, mat.DIAM_DRILL_ENG, 1* u.m),
                  (20 * u.L/u.s, 20 * u.cm, mat.DIAM_DRILL_ENG, 1* u.m),
                  (20 * u.L/u.s, 40 * u.cm, mat.DIAM_DRILL_ENG, 1* u.m))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(lfom.orifice_diameter(i[0], i[1], i[2]), i[3])

    def test_drillbit_area(self):
        """drillbit_area should give known result with known input.
        Test cases were calculated using outputs from original Mathcad code and
        the nominal diameter function written in pipedatabase which has already
        been tested.

        """
        checks = ((60 * u.L/u.s, 20 * u.cm, mat.DIAM_DRILL_ENG, 0.0007669903939428206 * u.m**2),
                  (60 * u.L/u.s, 40 * u.cm, mat.DIAM_DRILL_ENG, 0.0007669903939428206 * u.m**2),
                  (20 * u.L/u.s, 20 * u.cm, mat.DIAM_DRILL_ENG, 3.141592653589793 * u.m**2),
                  (20 * u.L/u.s, 40 * u.cm, mat.DIAM_DRILL_ENG, 0.0007669903939428206 * u.m**2))
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(lfom.drillbit_area(i[0], i[1], i[2]), i[3])




    def test_orifice_diameter(self):
        FLOW = 31 * u.L / u.s
        HL_LFOM = 20 * u.cm
        drill_bits = np.arange(5, 25, 5) * u.mm
        self.assertAlmostEqual(lfom.orifice_diameter(FLOW,HL_LFOM,drill_bits), 0.7874015748031495* u.m)


if __name__ == '__main__':
    unittest.main()



"""
Used to hold all constants related to AIDE. These should only be program specific constants and NOT
expert input constants (physical quantities). Those go in expert_inputs.
"""

DEFAULT_SIGNIFICANT_FIGURES = 4



import unittest
from aguaclara.core.units import unit_registry as u
from aguaclara.core import pipedatabase as pipe


class PipeTest(unittest.TestCase):
    def test_OD(self):
        checks = [[1.0 * u.inch, 1.315 * u.inch]]
        for i in checks:
            with self.subTest(i=i):
                self.assertAlmostEqual(pipe.OD(i[0]), i[1])

if __name__ == '__main__':
    unittest.main()



""" Minor Loss Coefficients of Reductions Module.

This module includes all minor loss coefficient calculations for common complex
geometries. For reductions and expansions, the following resource is used:
https://neutrium.net/fluid_flow/pressure-loss-from-fittings-expansion-and-reduction-in-pipe-size/

"""

from aguaclara.core.units import unit_registry as u
from aguaclara.core import physchem as pc
import aguaclara.core.constants as con
import aguaclara.core.materials_database as mats
import numpy as np
import aguaclara.core.utility as ut


@u.wraps(u.dimensionless, [u.m, u.m, u.L/u.s])
@ut.list_handler
def k_value_expansion(id_entrance:float, id_exit:float, flow, NU=con.NU_WATER, ROUGHNESS=mats.PIPE_ROUGH_PVC, theta=180, ROUNDED=False) -> float:
    """This function calculates the minor loss coefficient of a square, tapered or rounded expansion in a pipe
     using the equation defined here, where Re is the reynolds number on the inlet side, D_in and D_out are the inner diameter
     of the entrance pipe and exit pipe, respectively, and f_in is the Darcy friction factor of the inlet pipe:

    K_{square}  =   \left\{   {{    2\left[1-\left(\frac{D_{in}}{D_{out}}\right)^4\right]  if  Re_{in} < 4000}\atop
                   {     \left(1+0.8f_{in}\right)\left[1-\left(\frac{D_{in}}{D_{out}}\right)^{2}\right]^{2}  if  Re_{in} > 4000}}   \right.

    If the expansion is tapered, the K_{square} is multiplied by:

    C_{angle} =   \left\{   {{  1  if  45 < theta < 180}\atop
                             {   2.6\sin{\frac{\theta}{2}}  if  theta < 45}}   \right.

    If the expansion is rounded, k_{square} is used.

        Args:
            id_entrance (length): This is the inner pipe diameter of the pipe from which the fluid is flowing.
            id_exit (length): This is the inner pipe diameter of the pipe into which the fluid is flowing.
            flow (volume/time): This is the inner pipe diameter of the pipe into which the fluid is flowing.
            NU (float, optional): The fluid dynamic viscosity of the fluid. Defaults to water at room temperature (1 * 10**-6 * m**2/s)
            ROUGHNESS (length, optional): The pipe roughness. Defaults to PVC roughness.
            theta (angle, optional): The angle of the expansion. Defaults to a square reducer (180 degrees).
            ROUNDED (bool, optional): Whether or not the expansion is round. Defaults to square (False).

        Returns:
            k (float): the dimensionless minor loss coefficient, or k value of the expansion.
        """

    if id_entrance > id_exit:
        raise ValueError('For an expansion, the diameter of the exit pipe must be larger than that of the '
                         'entrance pipe. Use the k_value_reduction equation.')

    # Calculate constants
    # determine Darcy friction factor
    f = pc.fric(flow, id_entrance, NU, ROUGHNESS)
    # determine reynolds number in entrance pipe
    re = pc.re_pipe(flow, id_entrance, NU)

    # The reducer is rounded
    if ROUNDED:
        k = _k_value_rounded_expansion(id_entrance, id_exit, re)

    # The reducer is square
    elif theta == 180:
        k = _k_value_square_expansion(id_entrance, id_exit, re, f)

    # The reducer is tapered
    else:
        k = _k_value_tapered_expansion(id_entrance, id_exit, re, f)

    return k


@u.wraps(u.dimensionless, [u.m, u.m, u.L/u.s])
@ut.list_handler
def k_value_reduction(id_entrance:float, id_exit:float, flow, NU=con.NU_WATER, ROUGHNESS=mats.PIPE_ROUGH_PVC, theta=180, ROUNDED=False) -> float:
    """This function calculates the minor loss coefficient of a square, tapered or round reduction in a pipe
     using the equation defined here, where Re is the reynolds number on the inlet side, D_in and D_out are the inner diameter
     of the entrance pipe and exit pipe, respectively, and f_in is the Darcy friction factor of the inlet pipe:

    K_{square} =   \left\{   {{   \left(1.2+\frac{160}{Re_{in}}\right)\left[\left(\frac{D_{in}}{D_{out}}\right)^4-1\right]  if  Re_{in} < 2500}\atop
                              {   \left(0.6+0.48f_{in}\right)\left(\frac{D_{in}}{D_{out}}\right)^2\left[\left(\frac{D_{in}}{D_{out}}\right)^2-1\right]  if  Re_{in} > 2500}}   \right.

    If the reduction is tapered, the K_{square} is multiplied by:

    C_{angle} =   \left\{   {{  \sqrt{\sin{\frac{\theta}{2}}}}  if  45 < theta < 180}\atop
                             {   1.6\sin{\frac{\theta}{2}}  if  theta < 45}}   \right.

    If the reduction is rounded, it is calculated with the following equation:

    K_{rounded} = \left(0.1+\frac{50}{Re_{in}}\right)\left[\left(\frac{D_{in}}{D_{out}}\right)^4-1\right]



        Args:
            id_entrance (length): This is the inner pipe diameter of the pipe from which the fluid is flowing.
            id_exit (length): This is the inner pipe diameter of the pipe into which the fluid is flowing.
            flow (volume/time): This is the inner pipe diameter of the pipe into which the fluid is flowing.
            NU (float, optional): The fluid dynamic viscosity of the fluid. Defaults to water at room temperature (1 * 10**-6 * m**2/s)
            ROUGHNESS (length, optional): The pipe roughness. Defaults to PVC roughness.
            theta (angle, optional): The angle of the reducer. Defaults to a square reducer (180 degrees).
            ROUNDED (bool, optional): Whether or not the reducer is round. Defaults to square (False).

        Returns:
            k (float): the dimensionless minor loss coefficient, or k value of the reduction.
        """

    if id_entrance < id_exit:
        raise ValueError('For a reduction, the diameter of the exit pipe must be smaller than that of the '
                         'entrance pipe. Use the k_value_expansion equation.')

    # Calculate constants
    # determine Darcy friction factor
    f = pc.fric(flow, id_entrance, NU, ROUGHNESS)
    # determine reynolds number in entrance pipe
    re = pc.re_pipe(flow, id_entrance, NU)

    # The reducer is rounded
    if ROUNDED:
        k = _k_value_rounded_reduction(id_entrance, id_exit, re)

    # The reducer is square
    elif theta == 180:
        k = _k_value_square_reduction(id_entrance, id_exit, re, f)

    # The reducer is tapered
    else:
        k = _k_value_tapered_reduction(id_entrance, id_exit, re, f)

    return k


@u.wraps(u.dimensionless, [u.m, u.m, u.m, u.m**3/u.s])
@ut.list_handler
def k_value_orifice(id_pipe: float, id_orifice: float, length_orifice: float, flow: float, NU=con.NU_WATER) -> float:
    """This function calculates the minor loss coefficient of a thick and thin orifice plate in a pipe
     using the equation defined here, where Re is the reynolds number on the inlet side, and D_pipe and D_orifice are
     the inner diameter of the enclosing pipe and orifice, respectively, and L is the length of the orifice:

    K_{thin} =   \left\{   {{   \displaystyle K=\left[2.72+\left(\frac{D_{2}}{D_{1}}\right)^{2}\
                                                    left(\frac{120}{Re_{1}}-1\right)\right]\left[1-\
                                                    left(\frac{D_{2}}{D_{1}}\right)^{2}\right]\left[\
                                                    left(\frac{D_{1}}{D_{2}}\right)^{4}-1\right]
                                                    if  Re_{in} < 2500}\atop
                              {     \displaystyle K=\left[2.72+\left(\frac{D_{2}}{D_{1}}\right)^{2}\
                                                    left(\frac{4000}{Re_{1}}\right)\right]\left[1-\
                                                    left(\frac{D_{2}}{D_{1}}\right)^{2}\right]\left[\
                                                    left(\frac{D_{1}}{D_{2}}\right)^{4}-1\right]
                                                    if  Re_{in} > 2500}}   \right.

    If the reduction is longer so as not to be sharp, but still has a L/D_orifice < 5, k_thin is multiplied by the
    following coefficient:

    C_{length} =    \displaystyle  0.584+\left(\frac{0.0936}{\left(L/D_{2}\right)^{1.5}+0.225}\right)

    If the orifice is so long such that L/D_orifice > 5, a square reduction and expansion are used.

        Args:
            id_pipe (length): This is the inner pipe diameter of the pipe enclosing the orifice.
            id_orifice (length): This is the inner diameter of the orifice itself.
            length_orifice (length): This is the length of the orifice. 0 signifies a thin, sharp orifice.
            flow (volume/time): This is the inner pipe diameter of the pipe into which the fluid is flowing.
            NU (float, optional): The fluid dynamic viscosity of the fluid. Defaults to water at
                room temperature (1 * 10**-6 * m**2/s)
            ROUGHNESS (length, optional): The pipe roughness. Defaults to PVC roughness.

        Returns:
            k (float): the dimensionless minor loss coefficient (k value) of the orifice.
        """

    # The orifice cannot be larger than the enclosing pipe:
    if id_orifice > id_pipe:
        raise ValueError('The orifice cannot be larger than the enclosing pipe.')

    # Calculate constants
    # determine reynolds number in entrance pipe
    re = pc.re_pipe(flow, id_pipe, NU)

    # The orifice is thin
    if length_orifice == 0:
            k = _k_value_thin_sharp_orifice(id_pipe, id_orifice, re)

    # The orifice is thick
    elif length_orifice/id_orifice < 5:
        k = _k_value_thick_orifice(id_pipe, id_orifice, length_orifice, re)

    # The orifice is too thick to be considered an orifice, square reduction and expansion is used to model
    else:
        k = k_value_reduction(id_pipe * u.m, id_orifice * u.m, flow * u.m**3/u.s) + k_value_expansion(id_orifice * u.m,
                                                                                                      id_pipe * u.m,
                                                                                                      flow * u.m**3/u.s)
        # raise ValueError('For an orifice that is so long such that length_orifice/id_orifice > 5, use a'
        #                  ' reduction and expansion to determine the k value')

    return k

###############################################___________PRIVATE FUNCTIONS_____________################################

######### Reductions:


def _k_value_square_reduction(id_entrance: float, id_exit: float, re: float, f: float) -> float:
    # Calculate minor loss coefficient for square reducer
    if re < 2500:
        k = (1.2+(160/re))*((id_entrance/id_exit) ** 4)
    else:
        k = (0.6 + 0.48 * f) * (id_entrance / id_exit) ** 2 * ((id_entrance / id_exit) ** 2 - 1)
    return k


def _k_value_tapered_reduction(id_entrance: float, id_exit: float, re: float, f, theta: float) -> float:
    # Calculate minor loss coefficient for a tapered reducer
    k_square = _k_value_square_reduction(id_entrance, id_exit, re, f)
    if 45 < theta <= 180:
        k = k_square*np.sqrt(np.sin(theta / 2))
    elif 0 < theta <= 45:
        k = k_square * 1.6 * np.sin(theta / 2)
    else:
        raise ValueError('The reducer angle cannot be outside the [0,180] range')
    return k


def _k_value_rounded_reduction(id_entrance: float, id_exit: float, re: float) -> float:
    return (0.1 + (50 / re)) * ((id_entrance / id_exit) ** 4 - 1)

######### Expansions:


def _k_value_square_expansion(id_entrance: float, id_exit: float, re: float, f: float) -> float:
    # Calculate minor loss coefficient for square expansion
    if re < 4000:
        k = 2*(1-(id_entrance/id_exit) ** 4)
    else:
        k = (1 + 0.8 * f) * (1 - (id_entrance / id_exit) ** 2) ** 2
    return k


def _k_value_tapered_expansion(id_entrance: float, id_exit: float, re: float, f, theta: float) -> float:
    # Calculate minor loss coefficient for a tapered expansion
    k_square = _k_value_square_expansion(id_entrance, id_exit, re, f)
    if 45 < theta <= 180:
        k = k_square
    elif 0 < theta <= 45:
        k = k_square * 2.6 * np.sin(theta / 2)
    else:
        raise ValueError('The reducer angle cannot be outside the [0,180] range')
    return k

def _k_value_rounded_expansion(id_entrance: float, id_exit: float, re: float, f: float) -> float:
    return _k_value_square_expansion(id_entrance, id_exit, re, f)


######### Orifices:


def _k_value_thin_sharp_orifice(id_pipe: float, id_orifice: float, re: float) -> float:
    # Calculate minor loss coefficient for a thin, sharp orifice
    if re < 2500:
        k = ((2.72+(id_orifice/id_pipe) ** 2)*((120/re) - 1)) * (1 - (id_orifice/id_pipe) ** 2) * \
            ((id_pipe/id_orifice) ** 4 - 1)
    else:
        k = ((2.72 + (id_orifice / id_pipe) ** 2) * (4000 / re)) * (1 - (id_orifice / id_pipe) ** 2) * \
            ((id_pipe / id_orifice) ** 4 - 1)
    return k


def _k_value_thick_orifice(id_pipe: float, id_orifice: float, length_orifice: float, re: float) -> float:
    # Calculate minor loss coefficient for a thick orifice
    # if L/id_orifice > 5, use the equation for a square reduction and expansion.
    k_thin = _k_value_thin_sharp_orifice(id_pipe, id_orifice, re)
    # thickness coefficient:
    c = 0.584 + (0.0936 / ((length_orifice/id_orifice) ** 1.5 + 0.225))
    return c * k_thin



"""
research tests.
"""

import unittest
from aguaclara.research.utils import Constants, orbital_speed, circumference, orbital_period


class TestUtils(unittest.TestCase):
    """
    Test research's utils.
    """

    def setUp(self):
        pass

    def test_orbital_speed(self):
        """
        Calculate the orbital speed of an object.
        """
        answer = orbital_speed(
            Constants.Earth,
            600000,
            70
        )
        answer = round(answer, 3)
        self.assertEqual(
            answer,
            2425.552
        )

    def test_circumference(self):
        """
        2*pi*r
        """
        answer = circumference(600000)
        answer = round(answer, 3)
        self.assertEqual(
            answer,
            3769911.184
        )

    def test_orbital_period(self):
        """
        Calculate the orbital period of an object.
        """
        answer = orbital_period(
            Constants.Earth,
            600000,
            70
        )
        answer = round(answer, 3)
        self.assertEqual(
            answer,
            1554.43
        )




"""

"""

try:
    from aguaclara.core.units import unit_registry as u
    from aguaclara.core import constants as con, utility as ut
except ModuleNotFoundError:
    from aguaclara.core.units import unit_registry as u
    from aguaclara import utility as ut
    from aguaclara import constants as con


########### Materials Constants - general ############

PIPE_ROUGH_PVC = 0.12*u.mm

PIPE_ROUGH_CONCRETE = 2*u.mm

RHO_CONCRETE = 2400*(u.kg/(u.m**3)) #used in the sed tank drawing

THICKNESS_CONCRETE_MIN = 5*u.cm  #used throughout the code

#0 is English, 1 is metric, drill series is needed for the
#drill series at the bottom of this sheet and tube series
#is needed for the Cdc code
EN_DRILL_SERIES = 0

EN_TUBE_SERIES = 0

DIAM_REBAR = (1/2)*u.inch

########### Material constants - entrance tank ############

SDR_LFOM = 26

THICKNESS_LFOM_SHEET = THICKNESS_CONCRETE_MIN

NOM_DIAM_ENT_TANK_FLOAT =  8*u.inch

#Minimum pipe size to handle grit and to ensure that the pipe can be easily unclogged
NOM_DIAM_ENT_TANK_DRAIN_MIN =  3*u.inch

NOM_DIAM_ENT_TANK_DRAIN =  3*u.inch #This is constant for now

THICKNESS_ENT_TANK_REMOVABLE_WALL = 5*u.cm

#Parameters are arbitrary - need to be calculated
HEIGHT_ENT_TANK_REMOVABLE_WALL_SUPPORT = 4*u.cm

THICKNESS_ENT_TANK_REMOVABLE_WALL_SUPPORT = 5*u.cm

THICKNESS_ENT_TANK_HOPPER_LEDGE = 15*u.cm

THICKNESS_RAPID_MIX_ORIFICE_PLATE = 2*u.cm

NOM_DIAM_RAPID_MIX_AIR_RELEASE = 1*u.inch

 ############ Material constants - chem storage tanks  ############

THICKNESS_CHEM_TANK_WALL = 5*u.mm

 #Supplier Information:
#http://www.rotoplas.com/assets/files/industria/catalogo.pdf

#each element in the following array is tank volume

VOL_SUPPLIER_CHEM_TANK = [208.198, 450, 600, 750, 1100, 2500]*u.L

#the following array is a 2D array in which
#in each element, the first element is tank diameter
#and the second element is tank height

DIMENSIONS_SUPPLIER_CHEM_TANK = [[0.571, 0.851], [0.85, 0.99], [0.96, 1.10], [1.10, 1.02], [1.10, 1.39], [1.55, 1.65]]*u.m

FACTOR = [1.05, 1.05, 1.05, 1.05, 1.05, 1.05]

############ Material constants - chemical dose controller ###########

DIAM_TUBE_ENGLISH = [1, 2, 3 ,4 ,5 ,6, 7]*u.inch/16

DIAM_TUBE_METRIC = [2, 3, 4, 6, 8, 10]*u.mm

DIAM_FLT_VLV_ORIFICES_AVAIL = [0.093, 0.187, 0.25, 0.312]*u.inch

############# Material constants - flocculator #####################

THICKNESS_FLOC_BAFFLE_RIGID_HEIGHT = 15*u.cm

#The piping size for the main part of the floc modules
NOM_DIAM_FLOC_MODULES_MAIN = (1/2)*u.inch

#The diameter of the oversized cap used to assemble the floc modules
NOM_DIAM_FLOC_MODULES_LARGE = 1.5*u.inch

############ Material constants - sedimentation  #############
#Maximum length of sed plate sticking out past module pipes without any
#additional support. The goal is to prevent floppy modules that don't maintain
# constant distances between the plates

LENGTH_SED_PLATE_CANTILEVERED = 20*u.cm

NOM_DIAM_SED_HOPPER_DRAIN = 1*u.inch

NOM_DIAM_SED_HOPPER_VIEWER = 2*u.inch

NOM_DIAM_SED_HOPPER_SKIMMER = 2*u.inch

##Diffusers/Jet Reverser

NOM_DIAM_SED_DIFFUSER = 1*u.inch

NOM_DIAM_SED_JET_REVERSER = 3*u.inch

############ Material constants - stacked rapid sand filter ############

#We are going to take this pipe size for the slotted pipes as a given.
# Larger pipes may block too much flow and they are harder to install.
NOM_DIAM_FILTER_MANIFOLD_BRANCH = 1*u.inch

NOM_DIAM_FILTER_BACKWASH_MANIFOLD_BRANCH = 1.5*u.inch

#A slot thickness of 0.008 in or 0.2 mm is selected so that sand
# will not enter the slotted pipes.
WIDTH_FILTER_MANIFOLD_SLOTS = 0.008*u.inch

NOM_DIAM_FILTER_BRANCH_HOLDER = 2*u.inch

NOM_DIAM_FILTER_BACKWASH_BRANCH_HOLDER = 2*u.inch

#Minimum vertical spacing between trunk line pipes going through
#the filter wall for concrete construction
SPACE_FILTER_TRUNK_MIN = 3*u.cm

#Space between the ends of the branch receiver pipes and the walls so that
#the manifold assemblies are easy to lower into the filter boxes
# (if the branch receivers extended the entire length of the box they would
#just barely fit and it would be hard to get into place)
SPACE_FILTER_MANIFOLD_ASSEMBLY = 1*u.cm

LENGTH_FILTER_MANIFOLD_FEMCO_COUPLING = 4*u.cm

##Sand Properties

DIAM_FILTER_SAND_EFFECTIVE_SIZE = 0.5*u.mm

RATIO_UNIFORMITY_COEFF_FILTER_SAND = 1.65

DIAM_FILTER_SAND_60 = DIAM_FILTER_SAND_EFFECTIVE_SIZE * RATIO_UNIFORMITY_COEFF_FILTER_SAND

#Porosity in a sand bed
POROSITY_FILTER_SAND = 0.4

RHO_FILTER_SAND = 2650*u.kg/(u.m**3)

RATIO_FILTER_FLUIDIZED = 1.3 #Bed expands 30% when fluidized

#Carman-Kozeny coefficient
K_KOZENY = 5

############## Drill Series ##############

DIAM_DRILL_ENG = [0.03125, 0.0625, 0.09375, 0.125, 0.15625, 0.1875, 0.21875, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1, 1.25, 1.5, 1.75, 2]*u.inch

DIAM_DRILL_MET = [0.5*u.mm]

counter = 0
# these drill series should have been created using arange.

while DIAM_DRILL_MET[counter] <= 4.98*u.mm:
    counter+=1
    DIAM_DRILL_MET.append(DIAM_DRILL_MET[counter-1] + 0.1*u.mm)

while DIAM_DRILL_MET[counter] < 20*u.mm:
    counter+=1
    DIAM_DRILL_MET.append(DIAM_DRILL_MET[counter-1] + 1*u.mm)

while DIAM_DRILL_MET[counter] < 50*u.mm:
    counter+=1
    DIAM_DRILL_MET.append(DIAM_DRILL_MET[counter-1] + 2*u.mm)

def diam_drill(EN_DRILL_SERIES):
    if EN_DRILL_SERIES == 0:
        DIAM_DRILL = DIAM_DRILL_ENG
    else:
        DIAM_DRILL = DIAM_DRILL_MET
    return DIAM_DRILL



"""This file contains functions which convert between the nominal, inner, and
outer diameters of pipes based on their standard dimension ratio (SDR).

"""

#Let's begin to create the pipe database
# https://docs.python.org/2/library/csv.html
from aguaclara.core.units import unit_registry as u
import numpy as np
# We will use Pandas
import pandas as pd
# load the pipedb from a csv file

import os.path
dir_path = os.path.dirname(__file__)
csv_path = os.path.join(dir_path, 'data/pipedatabase.csv')
with open(csv_path) as pipedbfile:
    pipedb = pd.read_csv(pipedbfile)

@u.wraps(u.inch, u.inch, False)
def OD(ND):
    """Return a pipe's outer diameter according to its nominal diameter.

    The pipe schedule is not required here because all of the pipes of a
    given nominal diameter have the same outer diameter.

    Steps:
    1. Find the index of the closest nominal diameter.
       (Should this be changed to find the next largest ND?)
    2. Take the values of the array, subtract the ND, take the absolute
       value, find the index of the minimium value.
    """
    index = (np.abs(np.array(pipedb['NDinch']) - (ND))).argmin()
    return pipedb.iloc[index, 1]

@u.wraps(u.inch, [u.inch, None], False)
def ID_SDR(ND, SDR):
    """Return the inner diameter for SDR(standard diameter ratio) pipes.

    For these pipes the wall thickness is the outer diameter divided by
    the SDR.
    """
    return OD(ND).magnitude * (SDR-2) / SDR

@u.wraps(u.inch, u.inch, False)
def ID_sch40(ND):
    """Return the inner diameter for schedule 40 pipes.

    The wall thickness for these pipes is in the pipedb.

    Take the values of the array, subtract the ND, take the absolute
    value, find the index of the minimium value.
    """
    myindex = (np.abs(np.array(pipedb['NDinch']) - (ND))).argmin()
    return (pipedb.iloc[myindex, 1] - 2*(pipedb.iloc[myindex, 5]))


def ND_all_available():
    """Return an array of available nominal diameters.

    NDs available are those commonly used as based on the 'Used' column
    in the pipedb.
    """
    ND_all_available = []
    for i in range(len(pipedb['NDinch'])):
        if pipedb.iloc[i, 4] == 1:
            ND_all_available.append((pipedb['NDinch'][i]))
    return ND_all_available * u.inch


def ID_SDR_all_available(SDR):
    """Return an array of inner diameters with a given SDR.

    IDs available are those commonly used based on the 'Used' column
    in the pipedb.
    """
    ID = []
    ND = ND_all_available()
    for i in range(len(ND)):
        ID.append(ID_SDR(ND[i], SDR).magnitude)
    return ID * u.inch


def ND_SDR_available(ID, SDR):
    """ Return an available ND given an ID and a schedule.

    Takes the values of the array, compares to the ID, and finds the index
    of the first value greater or equal.
    """
    for i in range(len(np.array(ID_SDR_all_available(SDR)))):
        if np.array(ID_SDR_all_available(SDR))[i] >= (ID.to(u.inch)).magnitude:
            return ND_all_available()[i]


def ND_available(NDguess):
    """Return the minimum ND that is available.

    1. Extract the magnitude in inches from the nominal diameter.
    2. Find the index of the closest nominal diameter.
    3. Take the values of the array, subtract the ND, take the
       absolute value, find the index of the minimium value.
    """
    myindex = (ND_all_available() >= NDguess)
    return min(ND_all_available()[myindex])
